"""
Test the LS-Dyna parser.
"""

from lm_agent.models import LicenseUsesItem
from lm_agent.parsing.lsdyna import parse, parse_program_line, parse_total_line, parse_usage_line


def test_parse_program_line():
    """
    Does the regex for the program line match the lines in the output?
    The line contains:
    - program
    - expiration date
    - used CPUs
    - free CPUs
    - max CPUs
    - queue
    From these, we only need to extract ``program`` and ``max CPUs``.
    """
    assert parse_program_line("MPPDYNA          12/30/2022          0    500    500 |     0") == {
        "program": "mppdyna",
        "total": 500,
    }
    assert parse_program_line("MPPDYNA_971      12/30/2022          -     60    250 |     0") == {
        "program": "mppdyna_971",
        "total": 250,
    }
    assert parse_program_line("LS-DYNA          12/30/2022          0      0    500 |     0") == {
        "program": "ls-dyna",
        "total": 500,
    }
    assert parse_program_line("LS-DYNA_971      12/30/2022          0      0    500 |     0") == {
        "program": "ls-dyna_971",
        "total": 500,
    }
    assert parse_program_line("MPPDYNA          06/30/2024          -     60   2420 |     -") == {
        "program": "mppdyna",
        "total": 2420,
    }
    assert parse_program_line("not a program line") is None
    assert parse_program_line("MPPDYNA          12/30/2022          0") is None
    assert parse_program_line("") is None


def test_parse_usage_line():
    """
    Does the regex for the usage line match the lines in the output?
    The line contains:
    - user
    - port
    - host
    - used CPUs
    From these, we only need to extract ``user``, ``host``, ``used CPUs``.
    """
    assert parse_usage_line(" ywap0o     91665@n-c51.com  80") == LicenseUsesItem(
        username="ywap0o",
        lead_host="n-c51.com",
        booked=80,
    )
    assert parse_usage_line(" ndha1a     91665@n-c55.com  40") == LicenseUsesItem(
        username="ndha1a",
        lead_host="n-c55.com",
        booked=40,
    )
    assert parse_usage_line(" ywap0o     not-a-host  80") is None
    assert parse_usage_line(" ndha1a     91665@n-c55.com") is None
    assert parse_usage_line("91665@n-c55.com  20") is None
    assert parse_usage_line("") is None


def test_parse_total_line():
    """
    Does the regex for the total line match the line in the output?
    The line contains:
    - "LICENSE GROUP" string
    - used CPUs
    - free CPUs
    - max CPUs
    - queue
    From these, we only need to extract ``used CPUs``.
    """
    assert parse_total_line("                   LICENSE GROUP   440     60    500 |     0") == 440
    assert parse_total_line("                   LICENSE GROUP     0    000    500 |     0") == 0
    assert parse_total_line("                   0    000    500 |     0") is None
    assert parse_total_line("                   LICENSE GROUP     0") is None
    assert parse_total_line("") is None


def test_parse__correct_output(lsdyna_output):
    """
    Does the parser return the correct data for this output?
    - lsdyna_output: expected output from the license server,
    which contain licenses and usage information.
    """
    assert parse(lsdyna_output) == {
        "mppdyna": {
            "total": 500,
            "used": 440,
            "uses": [
                LicenseUsesItem(username="dvds3g", lead_host="n-c13.com", booked=80),
                LicenseUsesItem(username="ssss1d", lead_host="n-c52.com", booked=80),
                LicenseUsesItem(username="ssss1d", lead_host="n-c15.com", booked=80),
                LicenseUsesItem(username="ywap0o", lead_host="n-c53.com", booked=80),
                LicenseUsesItem(username="ywap0o", lead_host="n-c51.com", booked=80),
                LicenseUsesItem(username="ndha1a", lead_host="n-c55.com", booked=40),
            ],
        },
        "mppdyna_971": {"total": 500, "used": 440, "uses": []},
        "mppdyna_970": {"total": 500, "used": 440, "uses": []},
        "mppdyna_960": {"total": 500, "used": 440, "uses": []},
        "ls-dyna": {"total": 500, "used": 440, "uses": []},
        "ls-dyna_971": {"total": 500, "used": 440, "uses": []},
        "ls-dyna_970": {"total": 500, "used": 440, "uses": []},
        "ls-dyna_960": {"total": 500, "used": 440, "uses": []},
    }


def test_parse__bad_output(lsdyna_output_bad):
    """
    Does the parser return the correct data for this output?
    - lsdyna_output_bad: unparseable output from the license server,
    which can happen when a connection error occours.
    """
    assert parse(lsdyna_output_bad) == {}


def test_parse__no_licenses_output(lsdyna_output_no_licenses):
    """
    Does the parser return the correct data for this output?
    - lsdyna_output_no_licenses: expected output from the server
    when none of the licenses are in use by users.
    """
    assert parse(lsdyna_output_no_licenses) == {
        "mppdyna": {"total": 500, "used": 0, "uses": []},
        "mppdyna_971": {"total": 500, "used": 0, "uses": []},
        "mppdyna_970": {"total": 500, "used": 0, "uses": []},
        "mppdyna_960": {"total": 500, "used": 0, "uses": []},
        "ls-dyna": {"total": 500, "used": 0, "uses": []},
        "ls-dyna_971": {"total": 500, "used": 0, "uses": []},
        "ls-dyna_970": {"total": 500, "used": 0, "uses": []},
        "ls-dyna_960": {"total": 500, "used": 0, "uses": []},
    }



"""
Test the OLicense license server interface.
"""

from unittest import mock

from pytest import fixture, mark, raises

from lm_agent.config import settings
from lm_agent.exceptions import CommandFailedToExecute, LicenseManagerBadServerOutput
from lm_agent.models import LicenseReportItem
from lm_agent.server_interfaces.olicense import OLicenseLicenseServer


@fixture
def olicense_server(one_configuration_row_olicense) -> OLicenseLicenseServer:
    return OLicenseLicenseServer(one_configuration_row_olicense.license_servers)


def test_get_olicense_commands_list(olicense_server: OLicenseLicenseServer):
    """
    Do the commands for invoking the license server have the correct data?
    """
    commands_list = olicense_server.get_commands_list()
    assert commands_list == [
        [
            f"{settings.OLIXTOOL_PATH}",
            "-sv",
            "127.0.0.1:2345",
        ]
    ]


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.olicense.run_command")
async def test_olicense_get_output_from_server(
    run_command_mock: mock.MagicMock,
    olicense_server: OLicenseLicenseServer,
    olicense_output: str,
):
    """
    Do the license server interface return the output from the license server?
    """
    run_command_mock.return_value = olicense_output

    output = await olicense_server.get_output_from_server()
    assert output == olicense_output


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.olicense.OLicenseLicenseServer.get_output_from_server")
async def test_olicense_get_report_item(
    get_output_from_server_mock: mock.MagicMock, olicense_server: OLicenseLicenseServer, olicense_output: str
):
    """
    Do the OLicense server interface generate a report item for the product?
    """
    get_output_from_server_mock.return_value = olicense_output

    assert await olicense_server.get_report_item(1, "cosin.ftire_adams") == LicenseReportItem(
        feature_id=1,
        product_feature="cosin.ftire_adams",
        used=3,
        total=4,
        uses=[
            {"username": "sbhyma", "lead_host": "RD0087712", "booked": 1},
            {"username": "sbhyma", "lead_host": "RD0087713", "booked": 1},
            {"username": "user22", "lead_host": "RD0087713", "booked": 1},
        ],
    )


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.olicense.OLicenseLicenseServer.get_output_from_server")
async def test_olicense_get_report_item_with_bad_output(
    get_output_from_server_mock: mock.MagicMock,
    olicense_server: OLicenseLicenseServer,
    olicense_output_bad: str,
):
    """
    Do the OLicense server interface raise an exception when the server returns an unparseable output?
    """
    get_output_from_server_mock.return_value = olicense_output_bad

    with raises(LicenseManagerBadServerOutput):
        await olicense_server.get_report_item(1, "cosin.ftire_adams")


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.olicense.OLicenseLicenseServer.get_output_from_server")
async def test_olicense_get_report_item_with_no_used_licenses(
    get_output_from_server_mock: mock.MagicMock,
    olicense_server: OLicenseLicenseServer,
    olicense_output_no_licenses: str,
):
    """
    Do the OLicense server interface generate a report item when no licenses are in use?
    """
    get_output_from_server_mock.return_value = olicense_output_no_licenses

    assert await olicense_server.get_report_item(1, "cosin.ftire_adams") == LicenseReportItem(
        feature_id=1,
        product_feature="cosin.ftire_adams",
        used=0,
        total=4,
        used_licenses=[],
    )


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.olicense.run_command")
@mock.patch("lm_agent.server_interfaces.olicense.OLicenseLicenseServer.get_commands_list")
async def test_olicense_get_report_item_continues_on_exception(
    get_commands_list_mock: mock.MagicMock,
    run_command_mock: mock.MagicMock,
    olicense_server: OLicenseLicenseServer,
    olicense_output: str,
):
    """
    Do the OLicense server interface check the next command if the previous one fails?
    """
    get_commands_list_mock.return_value = [
        [
            f"{settings.OLIXTOOL_PATH}",
            "-sv",
            "127.0.0.1:2345",
        ],
        [
            f"{settings.OLIXTOOL_PATH}",
            "-sv",
            "127.0.1:3456",
        ],
    ]

    run_command_mock.side_effect = [
        CommandFailedToExecute("Command failed for first server"),
        olicense_output,
    ]

    assert await olicense_server.get_report_item(1, "cosin.ftire_adams") == LicenseReportItem(
        feature_id=1,
        product_feature="cosin.ftire_adams",
        used=3,
        total=4,
        uses=[
            {"username": "sbhyma", "lead_host": "RD0087712", "booked": 1},
            {"username": "sbhyma", "lead_host": "RD0087713", "booked": 1},
            {"username": "user22", "lead_host": "RD0087713", "booked": 1},
        ],
    )



"""
Test the DSLS license server interface.
"""

from unittest import mock

from pytest import fixture, mark, raises

from lm_agent.config import settings
from lm_agent.exceptions import CommandFailedToExecute, LicenseManagerBadServerOutput
from lm_agent.models import LicenseReportItem, LicenseUsesItem
from lm_agent.server_interfaces.dsls import DSLSLicenseServer


@fixture
def dsls_server(one_configuration_row_dsls) -> DSLSLicenseServer:
    return DSLSLicenseServer(one_configuration_row_dsls.license_servers)


def test_get_dsls_commands_list(dsls_server: DSLSLicenseServer):
    """
    Do the commands for invoking the license server have the correct data?
    """
    commands_list = dsls_server.get_commands_list()
    assert commands_list == [
        {
            "input": "connect 127.0.0.1 2345\ngetLicenseUsage -csv",
            "command": [f"{settings.DSLICSRV_PATH}", "-admin"],
        },
    ]


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.dsls.run_command")
async def test_dsls_get_output_from_server(
    run_command_mock: mock.MagicMock,
    dsls_server: DSLSLicenseServer,
    dsls_output: str,
):
    """
    Do the license server interface return the output from the license server?
    """
    run_command_mock.return_value = dsls_output

    output = await dsls_server.get_output_from_server()
    assert output == dsls_output


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.dsls.DSLSLicenseServer.get_output_from_server")
async def test_dsls_get_report_item(
    get_output_from_server_mock: mock.MagicMock, dsls_server: DSLSLicenseServer, dsls_output: str
):
    """
    Do the DSLS server interface generate a report item for the product?
    """
    get_output_from_server_mock.return_value = dsls_output

    assert await dsls_server.get_report_item(1, "powerflow.pw7") == LicenseReportItem(
        feature_id=1,
        product_feature="powerflow.pw7",
        used=2,
        total=2000,
        uses=[
            LicenseUsesItem(username="user_1", lead_host="nid001627", booked=2),
        ],
    )


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.dsls.DSLSLicenseServer.get_output_from_server")
async def test_dsls_get_report_item_with_bad_output(
    get_output_from_server_mock: mock.MagicMock, dsls_server: DSLSLicenseServer, dsls_output_bad: str
):
    """
    Do the DSLS server interface raise an exception when the server returns an unparseable output?
    """
    get_output_from_server_mock.return_value = dsls_output_bad

    with raises(LicenseManagerBadServerOutput):
        await dsls_server.get_report_item(1, "powerflow.pw7")


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.dsls.DSLSLicenseServer.get_output_from_server")
async def test_dsls_get_report_item_with_warning_message(
    get_output_from_server_mock: mock.MagicMock, dsls_server: DSLSLicenseServer, dsls_output_with_warning: str
):
    """
    Do the DSLS server interface parse the output when the warning line is present?
    """
    get_output_from_server_mock.return_value = dsls_output_with_warning

    assert await dsls_server.get_report_item(1, "powerflow.sru") == LicenseReportItem(
        feature_id=1,
        product_feature="powerflow.sru",
        used=493,
        total=2374,
        uses=[
            LicenseUsesItem(username="user_1", lead_host="nid001234", booked=493),
        ],
    )


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.dsls.DSLSLicenseServer.get_output_from_server")
async def test_dsls_get_report_item_with_no_used_licenses(
    get_output_from_server_mock: mock.MagicMock,
    dsls_server: DSLSLicenseServer,
    dsls_output_no_licenses: str,
):
    """
    Do the DSLSserver interface generate a report item when no licenses are in use?
    """
    get_output_from_server_mock.return_value = dsls_output_no_licenses

    assert await dsls_server.get_report_item(1, "powerflow.pw7") == LicenseReportItem(
        feature_id=1,
        product_feature="powerflow.pw7",
        used=0,
        total=2000,
        used_licenses=[],
    )


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.dsls.run_command")
@mock.patch("lm_agent.server_interfaces.dsls.DSLSLicenseServer.get_commands_list")
async def test_dsls_get_report_item_continues_on_exception(
    get_commands_list_mock: mock.MagicMock,
    run_command_mock: mock.MagicMock,
    dsls_server: DSLSLicenseServer,
    dsls_output: str,
):
    """
    Do the DSLS server interface check the next command if the previous one fails?
    """
    get_commands_list_mock.return_value = [
        {
            "input": "connect 127.0.0.1 2345\ngetLicenseUsage -csv",
            "command": [f"{settings.DSLICSRV_PATH}", "-admin"],
        },
        {
            "input": "connect 127.0.0.0 3456\ngetLicenseUsage -csv",
            "command": [f"{settings.DSLICSRV_PATH}", "-admin"],
        },
    ]
    run_command_mock.side_effect = [CommandFailedToExecute("Command failed for first server"), dsls_output]

    assert await dsls_server.get_report_item(1, "powerflow.pw7") == LicenseReportItem(
        feature_id=1,
        product_feature="powerflow.pw7",
        used=2,
        total=2000,
        uses=[
            LicenseUsesItem(username="user_1", lead_host="nid001627", booked=2),
        ],
    )



"""
Configuration of pytest for agent tests.
"""

from textwrap import dedent
from unittest.mock import patch

import httpx
import respx
from pytest import fixture

from lm_agent.config import settings
from lm_agent.models import (
    BookingSchema,
    ConfigurationSchema,
    FeatureSchema,
    JobSchema,
    LicenseReportItem,
    LicenseServerSchema,
    LicenseServerType,
    ProductSchema,
)


@fixture(autouse=True)
def mock_cache_dir(tmp_path):
    """Mock a cache directory."""
    _cache_dir = tmp_path / "license-manager-cache"
    assert not _cache_dir.exists()
    with patch("lm_agent.config.settings.CACHE_DIR", new=_cache_dir):
        yield _cache_dir


@fixture(autouse=True)
def mock_log_dir(tmp_path):
    """Mock a log directory."""
    _log_dir = tmp_path / "license-manager-agent-logs"
    assert not _log_dir.exists()
    with patch("lm_agent.config.settings.LOG_BASE_DIR", new=_log_dir):
        yield _log_dir


@fixture
def license_servers():
    """List of license servers."""
    return ["172.0.1.2 2345", "172.0.1.3 2345"]


@fixture
def respx_mock():
    """
    Run a test in the respx context (similar to respx decorator, but it's a fixture).

    Mocks the OIDC route used to secure a token.
    """
    with respx.mock as mock:
        respx.post(f"https://{settings.OIDC_DOMAIN}/protocol/openid-connect/token").mock(
            return_value=httpx.Response(status_code=200, json=dict(access_token="dummy-token"))
        )
        yield mock


@fixture
def jobs():
    """Some job response examples."""
    return [
        {
            "id": 1,
            "slurm_job_id": "123",
            "cluster_client_id": "dummy",
            "username": "user1",
            "lead_host": "host1",
            "bookings": [
                {"id": 1, "job_id": 1, "feature_id": 1, "quantity": 12},
                {"id": 2, "job_id": 1, "feature_id": 2, "quantity": 50},
            ],
        },
        {
            "id": 2,
            "slurm_job_id": "456",
            "cluster_client_id": "dummy",
            "username": "user2",
            "lead_host": "host2",
            "bookings": [
                {"id": 3, "job_id": 2, "feature_id": 4, "quantity": 15},
                {"id": 4, "job_id": 2, "feature_id": 7, "quantity": 25},
            ],
        },
        {
            "id": 3,
            "slurm_job_id": "789",
            "cluster_client_id": "dummy-2",
            "username": "user3",
            "lead_host": "host3",
            "bookings": [
                {"id": 14, "job_id": 6, "feature_id": 4, "quantity": 5},
                {"id": 15, "job_id": 6, "feature_id": 7, "quantity": 17},
            ],
        },
    ]


@fixture
def parsed_jobs():
    """Some parsed job response examples."""
    return [
        JobSchema(
            id=1,
            slurm_job_id="123",
            cluster_client_id="dummy",
            username="user1",
            lead_host="host1",
            bookings=[
                BookingSchema(id=1, job_id=1, feature_id=1, quantity=100),
                BookingSchema(id=2, job_id=1, feature_id=2, quantity=50),
            ],
        ),
        JobSchema(
            id=2,
            slurm_job_id="456",
            cluster_client_id="dummy",
            username="user2",
            lead_host="host2",
            bookings=[
                BookingSchema(id=3, job_id=2, feature_id=1, quantity=15),
                BookingSchema(id=4, job_id=2, feature_id=2, quantity=25),
            ],
        ),
        JobSchema(
            id=3,
            slurm_job_id="789",
            cluster_client_id="dummy",
            username="user3",
            lead_host="host3",
            bookings=[
                BookingSchema(id=14, job_id=6, feature_id=1, quantity=5),
                BookingSchema(id=15, job_id=6, feature_id=2, quantity=17),
            ],
        ),
    ]


@fixture
def one_parsed_job():
    """One parsed job response example."""
    return JobSchema(
        id=1,
        slurm_job_id="123",
        cluster_client_id="dummy",
        username="user1",
        lead_host="host1",
        bookings=[
            BookingSchema(id=1, job_id=1, feature_id=1, quantity=12),
            BookingSchema(id=2, job_id=1, feature_id=2, quantity=50),
        ],
    )


@fixture
def another_parsed_job():
    """Another parsed job response example."""
    return JobSchema(
        id=2,
        slurm_job_id="456",
        cluster_client_id="dummy",
        username="user2",
        lead_host="host2",
        bookings=[
            BookingSchema(id=3, job_id=2, feature_id=4, quantity=15),
            BookingSchema(id=4, job_id=2, feature_id=7, quantity=25),
        ],
    )


@fixture
def parsed_job_without_bookings():
    """One parsed job response example without bookings."""
    return JobSchema(
        id=1,
        slurm_job_id="789",
        cluster_client_id="dummy",
        username="user1",
        lead_host="host1",
        bookings=[],
    )


@fixture
def bookings():
    """Some booking response examples."""
    return [
        {"id": 1, "job_id": 1, "feature_id": 1, "quantity": 12},
        {"id": 2, "job_id": 1, "feature_id": 2, "quantity": 50},
        {"id": 3, "job_id": 2, "feature_id": 4, "quantity": 15},
        {"id": 4, "job_id": 2, "feature_id": 7, "quantity": 25},
        {"id": 14, "job_id": 6, "feature_id": 4, "quantity": 5},
        {"id": 15, "job_id": 6, "feature_id": 7, "quantity": 17},
    ]


@fixture
def configurations():
    """Some configuration response examples."""
    return [
        {
            "id": 1,
            "name": "Abaqus",
            "cluster_client_id": "dummy",
            "features": [
                {
                    "id": 1,
                    "name": "abaqus",
                    "product": {"id": 1, "name": "abaqus"},
                    "config_id": 1,
                    "reserved": 100,
                    "total": 123,
                    "used": 12,
                    "booked_total": 12,
                }
            ],
            "license_servers": [
                {"id": 1, "config_id": 1, "host": "licserv0001", "port": 1234},
                {"id": 3, "config_id": 1, "host": "licserv0003", "port": 8760},
            ],
            "grace_time": 60,
            "type": "flexlm",
        },
        {
            "id": 2,
            "name": "Converge",
            "cluster_client_id": "dummy",
            "features": [
                {
                    "id": 2,
                    "name": "converge_super",
                    "product": {"id": 2, "name": "converge"},
                    "config_id": 2,
                    "reserved": 0,
                    "total": 500,
                    "used": 50,
                    "booked_total": 50,
                }
            ],
            "license_servers": [{"id": 2, "config_id": 2, "host": "licserv0002", "port": 2345}],
            "grace_time": 123,
            "type": "rlm",
        },
    ]


@fixture
def parsed_configurations():
    """Some parsed configuration response examples."""
    return [
        ConfigurationSchema(
            id=1,
            name="Abaqus",
            cluster_client_id="dummy",
            features=[
                FeatureSchema(
                    id=1,
                    name="abaqus",
                    product=ProductSchema(id=1, name="abaqus"),
                    config_id=1,
                    reserved=100,
                    total=123,
                    used=12,
                    booked_total=12,
                )
            ],
            license_servers=[
                LicenseServerSchema(id=1, config_id=1, host="licserv0001", port=1234),
                LicenseServerSchema(id=3, config_id=1, host="licserv0003", port=8760),
            ],
            grace_time=60,
            type=LicenseServerType.FLEXLM,
        ),
        ConfigurationSchema(
            id=2,
            name="Converge",
            cluster_client_id="dummy",
            features=[
                FeatureSchema(
                    id=2,
                    name="converge_super",
                    product=ProductSchema(id=2, name="converge"),
                    config_id=2,
                    reserved=0,
                    total=500,
                    used=50,
                    booked_total=50,
                )
            ],
            license_servers=[LicenseServerSchema(id=2, config_id=2, host="licserv0002", port=2345)],
            grace_time=123,
            type=LicenseServerType.RLM,
        ),
    ]


@fixture
def features():
    """Some features reponse examples."""
    return [
        {
            "id": 1,
            "name": "abaqus",
            "product": {"id": 1, "name": "abaqus"},
            "config_id": 1,
            "reserved": 100,
            "total": 123,
            "used": 12,
            "booked_total": 12,
        },
        {
            "id": 2,
            "name": "abaqus",
            "product": {"id": 1, "name": "abaqus"},
            "config_id": 2,
            "reserved": 100,
            "total": 123,
            "used": 12,
            "booked_total": 50,
        },
    ]


@fixture
def parsed_features():
    """Some parsed features reponse examples."""
    return [
        FeatureSchema(
            id=1,
            name="abaqus",
            product=ProductSchema(id=1, name="abaqus"),
            config_id=1,
            reserved=100,
            total=123,
            used=12,
            booked_total=12,
        ),
        FeatureSchema(
            id=2,
            name="abaqus",
            product=ProductSchema(id=1, name="abaqus"),
            config_id=2,
            reserved=100,
            total=123,
            used=12,
            booked_total=50,
        ),
    ]


@fixture
def one_report_item():
    """Report item from a feature."""
    return LicenseReportItem(
        feature_id=1,
        product_feature="product.feature",
        used=100,
        total=1000,
        uses=[
            {"username": "user1", "lead_host": "host1", "booked": 100},
        ],
    )


@fixture
def another_report_item():
    """Another report item from a feature."""
    return LicenseReportItem(
        feature_id=2,
        product_feature="product2.feature2",
        used=10,
        total=100,
        uses=[
            {"username": "user2", "lead_host": "host2", "booked": 10},
        ],
    )


@fixture
def parsed_report_items():
    """Some parsed report items."""
    return [
        LicenseReportItem(
            feature_id=1,
            product_feature="abaqus.abaqus",
            used=100,
            total=1000,
            uses=[
                {"username": "user1", "lead_host": "host1", "booked": 100},
            ],
        ),
        LicenseReportItem(
            feature_id=2,
            product_feature="converge.converge_super",
            used=10,
            total=100,
            uses=[
                {"username": "user2", "lead_host": "host2", "booked": 10},
            ],
        ),
    ]


@fixture
def one_configuration_row_flexlm():
    return ConfigurationSchema(
        id=1,
        name="Test Feature",
        cluster_client_id="dummy",
        features=[
            FeatureSchema(
                id=1,
                name="testfeature",
                product=ProductSchema(id=1, name="testproduct"),
                config_id=1,
                reserved=100,
                total=1000,
                used=93,
                booked_total=0,
            )
        ],
        license_servers=[
            LicenseServerSchema(id=1, config_id=1, host="127.0.0.1", port=2345),
        ],
        grace_time=60,
        type=LicenseServerType.FLEXLM,
    )


@fixture
def one_configuration_row_rlm():
    return ConfigurationSchema(
        id=1,
        name="Converge",
        cluster_client_id="dummy",
        features=[
            FeatureSchema(
                id=1,
                name="converge_super",
                product=ProductSchema(id=1, name="converge"),
                config_id=1,
                reserved=100,
                total=1000,
                used=93,
                booked_total=0,
            )
        ],
        license_servers=[
            LicenseServerSchema(id=1, config_id=1, host="127.0.0.1", port=2345),
        ],
        grace_time=60,
        type=LicenseServerType.RLM,
    )


@fixture
def one_configuration_row_lsdyna():
    return ConfigurationSchema(
        id=1,
        name="MPPDYNA",
        cluster_client_id="dummy",
        features=[
            FeatureSchema(
                id=1,
                name="mppdyna",
                product=ProductSchema(id=1, name="mppdyna"),
                config_id=1,
                reserved=100,
                total=500,
                used=440,
                booked_total=0,
            )
        ],
        license_servers=[
            LicenseServerSchema(id=1, config_id=1, host="127.0.0.1", port=2345),
        ],
        grace_time=60,
        type=LicenseServerType.LSDYNA,
    )


@fixture
def one_configuration_row_lmx():
    return ConfigurationSchema(
        id=1,
        name="HyperWorks",
        cluster_client_id="dummy",
        features=[
            FeatureSchema(
                id=1,
                name="hyperworks",
                product=ProductSchema(id=1, name="hyperworks"),
                config_id=1,
                reserved=100,
                total=1000,
                used=93,
                booked_total=0,
            )
        ],
        license_servers=[
            LicenseServerSchema(id=1, config_id=1, host="127.0.0.1", port=2345),
        ],
        grace_time=60,
        type=LicenseServerType.LMX,
    )


@fixture
def one_configuration_row_olicense():
    return ConfigurationSchema(
        id=1,
        name="FTire Adams",
        cluster_client_id="dummy",
        features=[
            FeatureSchema(
                id=1,
                name="ftire_adams",
                product=ProductSchema(id=1, name="cosin"),
                config_id=1,
                reserved=100,
                total=1000,
                used=93,
                booked_total=0,
            )
        ],
        license_servers=[
            LicenseServerSchema(id=1, config_id=1, host="127.0.0.1", port=2345),
        ],
        grace_time=60,
        type=LicenseServerType.OLICENSE,
    )


@fixture
def one_configuration_row_dsls():
    return ConfigurationSchema(
        id=1,
        name="Powerflow-decomp",
        cluster_client_id="dummy",
        features=[
            FeatureSchema(
                id=1,
                name="pw7",
                product=ProductSchema(id=1, name="powerflow"),
                config_id=1,
                reserved=100,
                total=1000,
                used=93,
                booked_total=0,
            )
        ],
        license_servers=[
            LicenseServerSchema(id=1, config_id=1, host="127.0.0.1", port=2345),
        ],
        grace_time=60,
        type=LicenseServerType.DSLS,
    )


@fixture
def scontrol_show_lic_output_flexlm():
    """An output of scontrol show lic command for FlexLM license."""
    return dedent(
        """
        LicenseName=testproduct.testfeature@flexlm
            Total=10 Used=0 Free=10 Reserved=0 Remote=yes
        """
    )


@fixture
def scontrol_show_lic_output_rlm():
    """An output of scontrol show lic command for RLM license."""
    return dedent(
        """
        LicenseName=converge.converge_super@rlm
            Total=10 Used=0 Free=10 Reserved=0 Remote=yes
        """
    )


@fixture
def scontrol_show_lic_output_lsdyna():
    """An output of scontrol show lic command for LSDyna license."""
    return dedent(
        """
        LicenseName=mppdyna.mppdyna@lsdyna
            Total=500 Used=0 Free=500 Reserved=0 Remote=yes
        """
    )


@fixture
def scontrol_show_lic_output_lmx():
    """An output of scontrol show lic command for LM-X license."""
    return dedent(
        """
        LicenseName=hyperworks.hyperworks@lmx
            Total=1000000 Used=0 Free=500 Reserved=0 Remote=yes
        """
    )


@fixture
def scontrol_show_lic_output_olicense():
    """An output of scontrol show lic command for OLicense license."""
    return dedent(
        """
        LicenseName=cosin.ftire_adams@olicense
            Total=4 Used=0 Free=4 Reserved=0 Remote=yes
        """
    )


@fixture
def flexlm_output_bad():
    """Some unparseable lmstat output."""
    return dedent(
        """\
        lmstat - Copyright (c) 1989-2004 by Macrovision Corporation. All rights reserved.
        Flexible License Manager status on Wed 03/31/2021 09:12

        Error getting status: Cannot connect to license server (-15,570:111 "Connection refused")
        """
    )


@fixture
def flexlm_output():
    """Some FlexLM output to parse."""
    return dedent(
        """\
        lmstat - Copyright (c) 1989-2004 by Macrovision Corporation. All rights reserved.
        ...
        Users of TESTFEATURE:  (Total of 1000 licenses issued;  Total of 93 licenses in use)
        ...
            sdmfva myserver.example.com /dev/tty (v62.2) (myserver.example.com/24200 12507), start Thu 10/29 8:09, 29 licenses
            adfdna myserver.example.com /dev/tty (v62.2) (myserver.example.com/24200 12507), start Thu 10/29 8:09, 27 licenses
            sdmfva myserver.example.com /dev/tty (v62.2) (myserver.example.com/24200 12507), start Thu 10/29 8:09, 37 licenses
        """
    )


@fixture
def flexlm_output_2():
    """Some FlexLM output to parse."""
    return dedent(
        """\
        lmutil - Copyright (c) 1989-2012 Flexera Software LLC. All Rights Reserved.
        ...
        Users of TEST_FEATURE:  (Total of 42800 licenses issued;  Total of 1600 licenses in use)
        ...
            usbn12 p-c94.com /dev/tty feature=test_feature (v2023.0) (myserver.example.com/41020 10223), start Mon 3/11 13:16, 100 licenses
            usbn12 p-c94.com /dev/tty feature=test_feature_2 (v2023.0) (myserver.example.com/41020 626), start Mon 3/11 13:16, 1400 licenses
            usbn12 p-c94.com /dev/tty feature=test_feature_3 (v2023.0) (myserver.example.com/41020 10110), start Mon 3/11 13:16, 100 licenses
        """
    )


@fixture
def flexlm_output_3():
    """Some FlexLM output to parse."""
    return dedent(
        """\
        lmutil - Copyright (c) 1989-2012 Flexera Software LLC. All Rights Reserved.
        ...
        Users of ccmppower:  (Total of 40 licenses issued;  Total of 3 licenses in use)
        ...
        "ccmppower" v2025.09, vendor: abc
        floating license
            1nou7p dcv033.com /dev/tty (v2023.06) (myserver.example.com/27012 3457), start Mon 3/11 12:17

        "ccmppower" v2024.09, vendor: abc
        floating license

            1nou7p n-c41.com /dev/tty (v2023.06) (myserver.example.com/27012 2541), start Mon 3/11 22:36
            1nou7p nid001234 /dev/tty (v2022.10) (myserver.example.com/27012 3331), start Sun 3/10 10:41
        """
    )


@fixture
def flexlm_output_4():
    """Some FlexLM output to parse."""
    return dedent(
        """\
        lmutil - Copyright (c) 1989-2012 Flexera Software LLC. All Rights Reserved.
        ...
        Users of MSCONE:  (Total of 750 licenses issued;  Total of 18 licenses in use)
        ...
            ABCDKK ER0037 SESOR045 MSCONE:ADAMS_View (v2023.0331) (myserver.example.com/29065 2639), start Fri 3/8 13:25, 5 licenses
            ABCDKK ER0037 SESOR045 MSCONE:ADAMS_Car_Plugin (v2023.0331) (myserver.example.com/29065 8195), start Fri 3/8 13:25
            ABCDKK ER0037 SESOR045 MSCONE:ADAMS_View (v2023.0331) (myserver.example.com/29065 2474), start Fri 3/8 13:25, 5 licenses
            ABCDKK ER0037 SESOR045 MSCONE:ADAMS_Car_Plugin (v2023.0331) (myserver.example.com/29065 4903), start Fri 3/8 13:25
            ABCDKK ER0037 SESOR100 MSCONE:ADAMS_View (v2021.0630) (myserver.example.com/29065 11260), start Mon 3/11 10:49, 5 licenses
            ABCDKK ER0037 SESOR100 MSCONE:ADAMS_Car_Plugin (v2021.0630) (myserver.example.com/29065 7727), start Mon 3/11 10:49
        """
    )


@fixture
def flexlm_output_5():
    """Some FlexLM output to parse."""
    return dedent(
        """\
        Flexible License Manager status on Wed 7/3/2024 14:36
        ...
        Users of ADAMSTK_GG_Solver:  (Total of 6 licenses issued;  Total of 1 license in use)
        ...
            fjan1a dcv055.com /dev/pts/0 (v2023.0331) (myserver.example.com/29065 7549), start Wed 7/3 11:03
        """
    )


@fixture
def flexlm_output_no_licenses():
    """Some lmstat output with no licenses in use to parse."""
    return dedent(
        """\
        lmstat - Copyright (c) 1989-2004 by Macrovision Corporation. All rights reserved.
        ...

        Users of TESTFEATURE:  (Total of 1000 licenses issued;  Total of 0 licenses in use)

        ...

        """
    )


@fixture
def rlm_output_bad():
    """Some unparseable rlm output."""
    return dedent(
        """\
        rlmutil v12.2
        Copyright (C) 2006-2017, Reprise Software, Inc. All rights reserved.


        Error connecting to "rlm" server

        Connection attempted to host: "" on port 5053

        No error
        """
    )


@fixture
def rlm_output():
    """Some rlm output to parse."""
    return dedent(
        """\
        Setting license file path to 10@licserv.server.com
        rlmutil v12.2
        Copyright (C) 2006-2017, Reprise Software, Inc. All rights reserved.


            rlm status on licserv.com (port 35015), up 99d 11:08:25
            rlm software version v12.2 (build:2)
            rlm comm version: v1.2
            Startup time: Tue Oct 19 01:40:13 2021
            Todays Statistics (13:48:32), init time: Tue Nov  2 23:00:06 2021
            Recent Statistics (00:16:08), init time: Wed Nov  3 12:32:30 2021

                        Recent Stats         Todays Stats         Total Stats
                        00:16:08             13:48:32         15d 11:08:25
            Messages:    582 (0/sec)           28937 (0/sec)          777647 (0/sec)
            Connections: 463 (0/sec)           23147 (0/sec)          622164 (0/sec)

            --------- ISV servers ----------
            Name           Port Running Restarts
            csci             63133   Yes      0

            ------------------------

            csci ISV server status on licserv.server.com (port 63133), up 99d 11:08:18
            csci software version v12.2 (build: 2)
            csci comm version: v1.2
            csci Debug log filename: F:\RLM\Logs\csci.dlog
            csci Report log filename: F:\RLM\logs\Reportlogs\CSCILOG.rl
            Startup time: Tue Oct 19 01:40:20 2021
            Todays Statistics (13:48:32), init time: Tue Nov  2 23:00:06 2021
            Recent Statistics (00:16:08), init time: Wed Nov  3 12:32:30 2021

                        Recent Stats         Todays Stats         Total Stats
                        00:16:08             13:48:32         15d 11:08:18
            Messages:    991 (0/sec)           34770 (0/sec)          935961 (0/sec)
            Connections: 945 (0/sec)           17359 (0/sec)          466699 (0/sec)
            Checkouts:   0 (0/sec)           1 (0/sec)          937 (0/sec)
            Denials:     0 (0/sec)           0 (0/sec)          0 (0/sec)
            Removals:    0 (0/sec)           0 (0/sec)          0 (0/sec)


            ------------------------

            csci license pool status on licser.server.com (port 63133)

            converge v3.0
                count: 1, # reservations: 0, inuse: 0, exp: 31-jan-2022
                obsolete: 0, min_remove: 120, total checkouts: 0
            converge_gui v1.0
                count: 45, # reservations: 0, inuse: 0, exp: 31-jan-2022
                obsolete: 0, min_remove: 120, total checkouts: 26
            converge_gui_polygonica v1.0
                count: 1, # reservations: 0, inuse: 0, exp: 31-jan-2022
                obsolete: 0, min_remove: 120, total checkouts: 26
            converge_super v3.0
                count: 1000, # reservations: 0, inuse: 93, exp: 31-jan-2022
                obsolete: 0, min_remove: 120, total checkouts: 169
            converge_tecplot v1.0
                count: 45, # reservations: 0, inuse: 0, exp: 31-jan-2022
                obsolete: 0, min_remove: 120, total checkouts: 16


            ------------------------

            csci license usage status on licser.server.com (port 63133)

            converge_super v3.0: asdj13@myserver.example.com 29/0 at 11/01 09:01  (handle: 15a)
            converge_super v3.0: cddcp2@myserver.example.com 27/0 at 11/03 10:38  (handle: 128)
            converge_super v3.0: asdj13@myserver.example.com 37/0 at 11/01 09:01  (handle: 15a)
        """
    )


@fixture
def another_rlm_output():
    """Some rlm output to parse."""
    return dedent(
        """\
        Setting license file path to 1234@licserv0001.com
        rlmutil v12.2

            powercase v1.0
                count: 5, # reservations: 0, inuse: 3, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 19
            powerexport v1.0
                count: 5, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 0
            powerflow-decomp v1.0
                count: 3000, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 6
            powerflow-disc v1.0
                count: 3000, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 6
            powerinsight v1.0
                count: 5, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 5
            powerviz v1.0
                count: 5, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 26
            powerviz-generator v1.0
                count: 2, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 22
            powerviz-soiling v1.0
                count: 7, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 0
            exasignalprocessingjob v1.0
                count: 6, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 0
            poweracoustics v1.0
                count: 1, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 2
            powercool v1.0
                count: 1, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 0
            powerdelta v1.0
                count: 1, # reservations: 0, inuse: 1, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 9
            powerdelta-meshunion v1.0
                count: 1, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 0
            powerdelta-translate2 v1.0
                count: 1000000, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 0
            powerflow-sim v1.0
                count: 7000, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 6
            powerinsight-generator v1.0
                count: 1, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 0
            powerthermconnector v1.0
                count: 1, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 0

            ------------------------

            tainc license pool status on licserv0001.com (port 1234)

            powertherm v2023.1
                count: 1, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 0
            powertherm-mp v2023.1
                count: 3, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 0
            powertherm-multigrid v2023.1
                count: 1, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 0
            powertherm-sc v2023.1
                count: 1, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 0
            powertherm-solver v2023.1
                count: 1, # reservations: 0, inuse: 0, exp: 31-dec-2023
                obsolete: 0, min_remove: 120, total checkouts: 0


            ------------------------

            exacorp license usage status on licserv0001.com (port 1234)

            powercase v1.0: dfsdgv@server1 1/0 at 08/15 09:34  (handle: 182)
            powercase v1.0: addvbh@server2 1/0 at 08/16 09:36  (handle: c4)
            powercase v1.0: wrtgb3@server3 1/0 at 08/21 14:53  (handle: 11c)
            powerdelta v1.0: ghnds2@server4 1/0 at 08/23 15:16  (handle: a2)
        """
    )


@fixture
def rlm_output_no_licenses():
    """Some rlm output with no licenses in use to parse."""
    return dedent(
        """\
        Setting license file path to 35015@licserv0011.com:35015@licserv0012.com
        rlmutil v12.2
        Copyright (C) 2006-2017, Reprise Software, Inc. All rights reserved.


            rlm status on licserv0011.com (port 35015), up 20d 13:21:16
            rlm software version v12.2 (build:2)
            rlm comm version: v1.2
            Startup time: Tue Oct 19 03:40:13 2021
            Todays Statistics (16:01:23), init time: Mon Nov  8 00:00:06 2021
            Recent Statistics (00:28:35), init time: Mon Nov  8 15:32:54 2021

                         Recent Stats         Todays Stats         Total Stats
                         00:28:35              16:01:23         20d 13:21:16
            Messages:    997 (0/sec)           33562 (0/sec)          1033736 (0/sec)
            Connections: 797 (0/sec)           26849 (0/sec)          827039 (0/sec)

            --------- ISV servers ----------
               Name           Port Running Restarts
            csci             63133   Yes      0

            ------------------------

            csci ISV server status on licserv0011.com (port 63133), up 20d 13:21:09
            csci software version v12.2 (build: 2)
            csci comm version: v1.2
            csci Debug log filename: F:\RLM\Logs\csci.dlog
            csci Report log filename: F:\RLM\logs\Reportlogs\CSCILOG.rl
            Startup time: Tue Oct 19 03:40:20 2021
            Todays Statistics (16:01:23), init time: Mon Nov  8 00:00:06 2021
            Recent Statistics (00:28:35), init time: Mon Nov  8 15:32:54 2021
                         Recent Stats         Todays Stats         Total Stats
                          00:28:35             16:01:23         20d 13:21:09
            Messages:    1196 (0/sec)           40276 (0/sec)          1243764 (0/sec)
            Connections: 598 (0/sec)           20138 (0/sec)          620365 (0/sec)
            Checkouts:   0 (0/sec)           0 (0/sec)          262 (0/sec)
            Denials:     0 (0/sec)           0 (0/sec)          0 (0/sec)
            Removals:    0 (0/sec)           0 (0/sec)          0 (0/sec)


            ------------------------

            csci license pool status on licserv0011.com (port 63133)

            converge v3.0
                count: 1, # reservations: 0, inuse: 0, exp: 31-jan-2022
                obsolete: 0, min_remove: 120, total checkouts: 0
            converge_gui v1.0
                count: 45, # reservations: 0, inuse: 0, exp: 31-jan-2022
                obsolete: 0, min_remove: 120, total checkouts: 26
            converge_gui_polygonica v1.0
                count: 1, # reservations: 0, inuse: 0, exp: 31-jan-2022
                obsolete: 0, min_remove: 120, total checkouts: 26
            converge_super v3.0
                count: 1000, # reservations: 0, inuse: 0, exp: 31-jan-2022
                obsolete: 0, min_remove: 120, total checkouts: 189
            converge_tecplot v1.0
                count: 45, # reservations: 0, inuse: 0, exp: 31-jan-2022
                obsolete: 0, min_remove: 120, total checkouts: 21
        """
    )


@fixture
def lsdyna_output_bad():
    """Some unparseable lsdyna output."""
    return dedent(
        """\
        Using default server 31010@localhost
        *** ERROR failed to open server localhost
        """
    )


@fixture
def lsdyna_output():
    """Some lsdyna output to parse."""
    return dedent(
        """\
        Using user specified server 31010@licserv0004.com

        LICENSE INFORMATION

        PROGRAM          EXPIRATION CPUS  USED   FREE    MAX | QUEUE
        ---------------- ----------      ----- ------ ------ | -----
        MPPDYNA          12/30/2022          -     60    500 |     0
         dvds3g     59005@n-c13.com  80
         ssss1d     10274@n-c52.com  80
         ssss1d     47568@n-c15.com  80
         ywap0o    239793@n-c53.com  80
         ywap0o     91665@n-c51.com  80
         ndha1a     91626@n-c55.com  40
        MPPDYNA_971      12/30/2022          0     60    500 |     0
        MPPDYNA_970      12/30/2022          0     60    500 |     0
        MPPDYNA_960      12/30/2022          0     60    500 |     0
        LS-DYNA          12/30/2022          0     60    500 |     0
        LS-DYNA_971      12/30/2022          0     60    500 |     0
        LS-DYNA_970      12/30/2022          0     60    500 |     0
        LS-DYNA_960      12/30/2022          0     60    500 |     0
                           LICENSE GROUP   440     60    500 |     0
        """
    )


@fixture
def lsdyna_output_no_licenses():
    """Some lsdyna output with no licenses in use to parse."""
    return dedent(
        """\
        Using user specified server 31010@licserv0004.com

        LICENSE INFORMATION

        PROGRAM          EXPIRATION CPUS  USED   FREE    MAX | QUEUE
        ---------------- ----------      ----- ------ ------ | -----
        MPPDYNA          12/30/2022          0    500    500 |     0
        MPPDYNA_971      12/30/2022          0    500    500 |     0
        MPPDYNA_970      12/30/2022          0    000    500 |     0
        MPPDYNA_960      12/30/2022          0    000    500 |     0
        LS-DYNA          12/30/2022          0    000    500 |     0
        LS-DYNA_971      12/30/2022          0    000    500 |     0
        LS-DYNA_970      12/30/2022          0    000    500 |     0
        LS-DYNA_960      12/30/2022          0    000    500 |     0
                           LICENSE GROUP     0    000    500 |     0
        """
    )


@fixture
def lmx_output_bad():
    """Some unparseable LM-X output."""
    return dedent(
        """\
        LM-X End-user Utility v3.32
        Copyright (C) 2002-2010 X-Formation. All rights reserved.

        ++++++++++++++++++++++++++++++++++++++++
        LM-X license server(s):
        ----------------------------------------
        There are no license server(s) available.
        """
    )


@fixture
def lmx_output():
    """Some LM-X output to parse."""
    return dedent(
        """\
        LM-X End-user Utility v3.32
        Copyright (C) 2002-2010 X-Formation. All rights reserved.

        ++++++++++++++++++++++++++++++++++++++++
        LM-X License Server on 6200@licserv0013.com:

        Server version: v5.1 Uptime: 3 day(s) 12 hour(s) 0 min(s) 51 sec(s)
        ----------------------------------------
        Feature: CatiaV5Reader Version: 21.0 Vendor: ALTAIR
        Start date: 2022-02-17 Expire date: 2023-01-31
        Key type: EXCLUSIVE License sharing: CUSTOM VIRTUAL

        0 of 3 license(s) used
        ----------------------------------------
        Feature: GlobalZoneEU Version: 21.0 Vendor: ALTAIR
        Start date: 2022-02-17 Expire date: 2023-01-31
        Key type: EXCLUSIVE License sharing: CUSTOM VIRTUAL

        40000 of 1000003 license(s) used:

        15000 license(s) used by VRAAFG@RD0082879 [138.106.159.158]
        Login time: 2022-02-18 09:26   Checkout time: 2022-02-18 09:29
        Shared on custom string: VRAAFG:RD0082879

        25000 license(s) used by VRAAFG@RD0082879 [138.106.159.158]
        Login time: 2022-02-18 09:26   Checkout time: 2022-02-18 09:26
        Shared on custom string: VRAAFG:RD0082879
        ----------------------------------------
        Feature: HWAIFPBS Version: 21.0 Vendor: ALTAIR
        Start date: 2022-02-17 Expire date: 2023-01-31
        Key type: EXCLUSIVE License sharing: CUSTOM VIRTUAL

        0 of 2147483647 license(s) used
        ----------------------------------------
        Feature: HWAWPF Version: 21.0 Vendor: ALTAIR
        Start date: 2022-02-17 Expire date: 2023-01-31
        Key type: EXCLUSIVE License sharing: CUSTOM VIRTUAL

        0 of 2147483647 license(s) used
        ----------------------------------------
        Feature: HWActivate Version: 21.0 Vendor: ALTAIR
        Start date: 2022-02-17 Expire date: 2023-01-31
        Key type: EXCLUSIVE License sharing: CUSTOM VIRTUAL

        0 of 2147483647 license(s) used
        ----------------------------------------
        Feature: HWFlux2D Version: 21.0 Vendor: ALTAIR
        Start date: 2022-02-17 Expire date: 2023-01-31
        Key type: EXCLUSIVE License sharing: CUSTOM VIRTUAL

        30000 of 2147483647 license(s) used:

        15000 license(s) used by VRAAFG@RD0082879 [138.106.159.158]
        Login time: 2022-02-18 09:26   Checkout time: 2022-02-18 09:29
        Shared on custom string: VRAAFG:RD0082879:27164_23514544_1645091752_138525

        15000 license(s) used by VRAAFG@RD0082879 [138.106.159.158]
        Login time: 2022-02-18 09:26   Checkout time: 2022-02-18 09:26
        Shared on custom string: VRAAFG:RD0082879:18896_1081950704_1645017269_309963
        ----------------------------------------
        Feature: HyperWorks Version: 21.0 Vendor: ALTAIR
        Start date: 2022-02-17 Expire date: 2023-01-31
        Key type: EXCLUSIVE License sharing: CUSTOM VIRTUAL

        25000 of 1000000 license(s) used:

        25000 license(s) used by sssaah@RD0082406 [138.106.154.220]
        Login time: 2022-02-18 09:26   Checkout time: 2022-02-18 09:26
        Shared on custom string: sssaah:RD0082406
        """
    )


@fixture
def lmx_output_2():
    return dedent(
        """
        LM-X End-user Utility v3.32
        Copyright (C) 2002-2010 X-Formation. All rights reserved.

        ++++++++++++++++++++++++++++++++++++++++
        LM-X License Server on 6300@licserv0003.scom:

        Server version: v4.9.3 Uptime: 10 day(s) 16 hour(s) 54 min(s) 21 sec(s)
        ----------------------------------------
        Feature: FEMFAT_VISUALIZER Version: 2024.0 Vendor: abc
        Start date: NONE Expire date: 2024-06-30
        Key type: EXCLUSIVE License sharing: HOST USER VIRTUAL

        2 of 2 license(s) used:

        1 license(s) used by fdsva1@dcv046.com_ver2023 [10.123.321.20]
            Login time: 2024-03-11 12:50   Checkout time: 2024-03-11 12:50 
            Shared on username: fdsva1   Shared on hostname: dcv046.com_ver2023 

        1 license(s) used by asdsc1@dcv048.com_ver2022a [10.123.321.10]
            Login time: 2024-03-11 17:29   Checkout time: 2024-03-11 17:29 
            Shared on username: asdsc1   Shared on hostname: dcv048.com_ver2022a 
        """
    )


@fixture
def lmx_output_no_licenses():
    """Some LM-X output with no licenses in use to parse."""
    return dedent(
        """\
        LM-X End-user Utility v3.32
        Copyright (C) 2002-2010 X-Formation. All rights reserved.

        ++++++++++++++++++++++++++++++++++++++++
        LM-X License Server on 6200@licserv0013.com:

        Server version: v5.1 Uptime: 3 day(s) 12 hour(s) 0 min(s) 51 sec(s)
        ----------------------------------------
        Feature: CatiaV5Reader Version: 21.0 Vendor: ALTAIR
        Start date: 2022-02-17 Expire date: 2023-01-31
        Key type: EXCLUSIVE License sharing: CUSTOM VIRTUAL

        0 of 3 license(s) used
        ----------------------------------------
        Feature: GlobalZoneEU Version: 21.0 Vendor: ALTAIR
        Start date: 2022-02-17 Expire date: 2023-01-31
        Key type: EXCLUSIVE License sharing: CUSTOM VIRTUAL

        0 of 1000003 license(s) used
        ----------------------------------------
        Feature: HWAIFPBS Version: 21.0 Vendor: ALTAIR
        Start date: 2022-02-17 Expire date: 2023-01-31
        Key type: EXCLUSIVE License sharing: CUSTOM VIRTUAL

        0 of 2147483647 license(s) used
        ----------------------------------------
        Feature: HWAWPF Version: 21.0 Vendor: ALTAIR
        Start date: 2022-02-17 Expire date: 2023-01-31
        Key type: EXCLUSIVE License sharing: CUSTOM VIRTUAL

        0 of 2147483647 license(s) used
        ----------------------------------------
        Feature: HWActivate Version: 21.0 Vendor: ALTAIR
        Start date: 2022-02-17 Expire date: 2023-01-31
        Key type: EXCLUSIVE License sharing: CUSTOM VIRTUAL

        0 of 2147483647 license(s) used
        ----------------------------------------
        Feature: HWFlux2D Version: 21.0 Vendor: ALTAIR
        Start date: 2022-02-17 Expire date: 2023-01-31
        Key type: EXCLUSIVE License sharing: CUSTOM VIRTUAL

        0 of 2147483647 license(s) used
        ----------------------------------------
        Feature: HyperWorks Version: 21.0 Vendor: ALTAIR
        Start date: 2022-02-17 Expire date: 2023-01-31
        Key type: EXCLUSIVE License sharing: CUSTOM VIRTUAL

        0 of 1000000 license(s) used
        """
    )


@fixture
def olicense_output_bad():
    """Some OLicense unparseable  output."""
    return dedent(
        """\
        olixtool 4.8.0 - OLicense XML Status Application
        Copyright (C) 2007-2013 Optimum GmbH, Karlsruhe, Germany

        Server: <123:80:0> (Proxy: <>)

        OlComm 454 Request cannot connect to target
        """
    )


@fixture
def olicense_output():
    """Some OLicense output to parse."""
    return dedent(
        """\
        olixtool 4.8.0 - OLicense XML Status Application
        Copyright (C) 2007-2013 Optimum GmbH, Karlsruhe, Germany

        Server: <138.106.32.97:31212:0> (Proxy: <>)

        List of requested licenses:

        ==============================================
        Application:	cosin
        VersionRange:	0-20231
        Licenser:	cosin scientific software
        Licensee:	Tomas Fjällström
        License-ID:	cosin-faf71fdc@cos550218
        Modules:
          Name; LicenseType; FloatCount; Expiration
          --------------------------------------------
          ftire_adams;         	FreeFloating;	3;	2022-12-31 23:59:59;
            3 FloatsLockedBy:
              sbhyma@RD0087712 #1
              sbhyma@RD0087713 #1
              user22@RD0087713 #1


        ==============================================
        Application:	cosin
        VersionRange:	0-20231
        Licenser:	cosin scientific software
        Licensee:	Tomas Fjällström
        License-ID:	cosin-faf71fdc@cos558164
        Modules:
          Name; LicenseType; FloatCount; Expiration
          --------------------------------------------
          ftire_adams;         	FreeFloating;	1;	2023-02-28 23:59:00;
        """
    )


@fixture
def olicense_output_no_licenses():
    """Some OLicense output with no licenses in use to parse."""
    return dedent(
        """\
        olixtool 4.8.0 - OLicense XML Status Application
        Copyright (C) 2007-2013 Optimum GmbH, Karlsruhe, Germany

        Server: <138.106.32.97:31212:0> (Proxy: <>)

        List of requested licenses:

        ==============================================
        Application:	cosin
        VersionRange:	0-20231
        Licenser:	cosin scientific software
        Licensee:	Tomas Fjällström
        License-ID:	cosin-faf71fdc@cos550218
        Modules:
          Name; LicenseType; FloatCount; Expiration
          --------------------------------------------
          ftire_adams;         	FreeFloating;	3;	2022-12-31 23:59:59;

        ==============================================
        Application:	cosin
        VersionRange:	0-20231
        Licenser:	cosin scientific software
        Licensee:	Tomas Fjällström
        License-ID:	cosin-faf71fdc@cos558164
        Modules:
          Name; LicenseType; FloatCount; Expiration
          --------------------------------------------
          ftire_adams;         	FreeFloating;	1;	2023-02-28 23:59:00;
        """
    )


@fixture
def dsls_output():
    """
    Some DSLS output to parse.
    """
    return dedent(
        """\
        License Administration Tool Version 6.425.4 Built on May 2, 2023, 6:45:36 PM.
        admin >	Software version: 6.425.4
            Build date: May 2, 2023, 6:45:36 PM
            Standalone mode
            Ready: yes
            Server name: localhost   Server id: ZSD-123
        Editor,EditorId,Feature,Model,Commercial Type,Max Release Number,Max Release Date,Pricing Structure,Max Casual Duration,Expiration Date,Customer ID,Count,Inuse,Tokens,Casual Usage (mn),Host,User,Internal ID,Active Process,Client Code Version,Session ID,Granted Since,Last Used At,Granted At,Queue Position,
        Dassault Systemes,5E756A80,PAC,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,10001723,1,0,
        Dassault Systemes,5E756A80,PAJ,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,10001723,6,0,
        Dassault Systemes,5E756A80,PCA,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,10001723,4,0,
        Dassault Systemes,5E756A80,PCO,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,10001723,2,0,
        Dassault Systemes,5E756A80,PD5,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,10001723,1,0,
        Dassault Systemes,5E756A80,PV6,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,10001723,7,0,
        Dassault Systemes,5E756A80,PVG,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,10001723,2,0,
        Dassault Systemes,5E756A80,PW7,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,10001723,2000,2,2,,nid001627 (263.0)/127.0.0.1,user_1,PW7,/powerflow/pf_sim_comm ( 3148728),6.424,02DBE16,2024-09-17 17:59:34,2024-09-18 15:25:50,2024-09-17 17:59:34,
        Dassault Systemes,5E756A80,PW8,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,10001723,2000,0,
        Dassault Systemes,5E756A80,SRU,Token,STD,423,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,10001723,2374,1559,493,,nid001626 (263.0)/127.0.0.1,user_2,SRU,/powerflow/pf_sim_comm ( 3148728),6.424,02DBE16,2024-09-17 17:59:34,2024-09-18 15:25:50,2024-09-17 17:59:34,
        Dassault Systemes,5E756A80,SRU,Token,STD,423,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,10001723,2374,1559,533,,nid001601 (226.0)/127.0.0.1,user_3,SRU,/powerflow/pf_sim_comm ( 2885345),6.424,02DBE17,2024-09-17 18:48:17,2024-09-18 15:26:13,2024-09-17 18:48:17,
        Dassault Systemes,5E756A80,SRU,Token,STD,423,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,10001723,2374,1559,533,,nid001671 (2a8.0)/127.0.0.1,user_4,SRU,/powerflow/pf_sim_comm ( 3206577),6.424,02DBE18,2024-09-17 18:49:49,2024-09-18 15:26:45,2024-09-17 18:49:49,
        """
    )


@fixture
def dsls_output_no_licenses():
    """Some DSLS output with no licenses in use to parse."""
    return dedent(
        """\
        License Administration Tool Version 6.425.4 Built on May 2, 2023, 6:45:36 PM.
        admin >	Software version: 6.425.4
            Build date: May 2, 2023, 6:45:36 PM
            Standalone mode
            Ready: yes
            Server name: localhost   Server id: ZSD-123
        Editor,EditorId,Feature,Model,Commercial Type,Max Release Number,Max Release Date,Pricing Structure,Max Casual Duration,Expiration Date,Customer ID,Count,Inuse,Tokens,Casual Usage (mn),Host,User,Internal ID,Active Process,Client Code Version,Session ID,Granted Since,Last Used At,Granted At,Queue Position,
        Dassault Systemes,5E756A80,PAC,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,1001723,1,0,
        Dassault Systemes,5E756A80,PAJ,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,1001723,6,0,
        Dassault Systemes,5E756A80,PCA,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,1001723,4,0,
        Dassault Systemes,5E756A80,PCO,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,1001723,2,0,
        Dassault Systemes,5E756A80,PD5,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,1001723,1,0,
        Dassault Systemes,5E756A80,PV6,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,1001723,7,0,
        Dassault Systemes,5E756A80,PVG,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,1001723,2,0,
        Dassault Systemes,5E756A80,PW7,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,1001723,2000,0,
        Dassault Systemes,5E756A80,PW8,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,1001723,2000,0,
        Dassault Systemes,5E756A80,SRU,Token,STD,423,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,1001723,2374,0,
        """
    )


@fixture
def dsls_output_with_warning():
    """Some DSLS output with a warning message to parse."""
    return dedent(
        """\
        License Administration Tool Version 6.425.4 Built on May 2, 2023, 6:45:36 PM.
        admin >	Software version: 6.425.4
            Build date: May 2, 2023, 6:45:36 PM
            Standalone mode
            Ready: yes
            Server name: localhost   Server id: ZSD-123
        Warning: restricted connection, some operations are not permitted; due to an existing connection in full access from localhost(127.0.0.1)
        admin >Editor,EditorId,Feature,Model,Commercial Type,Max Release Number,Max Release Date,Pricing Structure,Max Casual Duration,Expiration Date,Customer ID,Count,Inuse,Tokens,Casual Usage (mn),Host,User,Internal ID,Active Process,Client Code Version,Session ID,Granted Since,Last Used At,Granted At,Queue Position,
        Dassault Systemes,5E756A80,PAC,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,1,0,
        Dassault Systemes,5E756A80,PAJ,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,6,0,
        Dassault Systemes,5E756A80,PCA,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,4,0,
        Dassault Systemes,5E756A80,PCO,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,2,0,
        Dassault Systemes,5E756A80,PD5,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,1,0,
        Dassault Systemes,5E756A80,PD6,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,1,0,
        Dassault Systemes,5E756A80,PD8,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,10000,0,
        Dassault Systemes,5E756A80,PEX,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,5,0,
        Dassault Systemes,5E756A80,PIG,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,1,0,
        Dassault Systemes,5E756A80,PIN,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,5,0,
        Dassault Systemes,5E756A80,PT5,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,2,0,
        Dassault Systemes,5E756A80,PT6,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,2,0,
        Dassault Systemes,5E756A80,PTC,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,2,0,
        Dassault Systemes,5E756A80,PTI,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,2,0,
        Dassault Systemes,5E756A80,PTL,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,6,0,
        Dassault Systemes,5E756A80,PTS,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,2,0,
        Dassault Systemes,5E756A80,PV5,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,5,0,
        Dassault Systemes,5E756A80,PV6,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,7,0,
        Dassault Systemes,5E756A80,PVG,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,2,0,
        Dassault Systemes,5E756A80,PW7,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,2000,0,
        Dassault Systemes,5E756A80,PW8,Token,STD,8,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,2000,0,
        Dassault Systemes,5E756A80,SRU,Token,STD,423,2025-01-01 00:59:00,YLC,0,2025-01-01 00:59:00,100000000001723,2374,493,493,,nid001234 (263.0)/127.0.0.1,user_1,SRU,,,,2024-12-09 03:02:19,2024-12-10 02:21:33,,
        """
    )


@fixture
def dsls_output_bad():
    """Some unparseable DSLS output."""
    return dedent(
        """\
        License Administration Tool Version 6.425.4 Built on May 2, 2023, 6:45:36 PM.
        Cannot connect to server localhost on port 1234
        """
    )



"""
Test the FlexLM license server interface.
"""

from unittest import mock

from pytest import fixture, mark, raises

from lm_agent.config import settings
from lm_agent.exceptions import CommandFailedToExecute, LicenseManagerBadServerOutput
from lm_agent.models import LicenseReportItem
from lm_agent.server_interfaces.flexlm import FlexLMLicenseServer


@fixture
def flexlm_server(one_configuration_row_flexlm) -> FlexLMLicenseServer:
    return FlexLMLicenseServer(one_configuration_row_flexlm.license_servers)


def test_get_flexlm_commands_list(flexlm_server: FlexLMLicenseServer):
    """
    Do the commands for invoking the license server have the correct data?
    """
    commands_list = flexlm_server.get_commands_list()
    assert commands_list == [
        [
            f"{settings.LMUTIL_PATH}",
            "lmstat",
            "-c",
            "2345@127.0.0.1",
            "-f",
        ]
    ]


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.flexlm.run_command")
async def test_flexlm_get_output_from_server(
    run_command_mock: mock.MagicMock,
    flexlm_server: FlexLMLicenseServer,
    flexlm_output: str,
):
    """
    Do the license server interface return the output from the license server?
    """
    run_command_mock.return_value = flexlm_output

    output = await flexlm_server.get_output_from_server("testproduct.testfeature")
    assert output == flexlm_output


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.flexlm.FlexLMLicenseServer.get_output_from_server")
async def test_flexlm_get_report_item(
    get_output_from_server_mock: mock.MagicMock, flexlm_server: FlexLMLicenseServer, flexlm_output: str
):
    """
    Do the FlexLM server interface generate a report item for the product?
    """
    get_output_from_server_mock.return_value = flexlm_output

    assert await flexlm_server.get_report_item(1, "testproduct.testfeature") == LicenseReportItem(
        feature_id=1,
        product_feature="testproduct.testfeature",
        used=93,
        total=1000,
        uses=[
            {"username": "sdmfva", "lead_host": "myserver.example.com", "booked": 29},
            {"username": "adfdna", "lead_host": "myserver.example.com", "booked": 27},
            {"username": "sdmfva", "lead_host": "myserver.example.com", "booked": 37},
        ],
    )


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.flexlm.FlexLMLicenseServer.get_output_from_server")
async def test_flexlm_get_report_item_with_bad_output(
    get_output_from_server_mock: mock.MagicMock, flexlm_server: FlexLMLicenseServer, flexlm_output_bad: str
):
    """
    Do the FlexLM server interface raise an exception when the server returns an unparseable output?
    """
    get_output_from_server_mock.return_value = flexlm_output_bad

    with raises(LicenseManagerBadServerOutput):
        await flexlm_server.get_report_item(1, "testproduct.testfeature")


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.flexlm.FlexLMLicenseServer.get_output_from_server")
async def test_flexlm_get_report_item_with_no_used_licenses(
    get_output_from_server_mock: mock.MagicMock,
    flexlm_server: FlexLMLicenseServer,
    flexlm_output_no_licenses: str,
):
    """
    Do the FlexLM server interface generate a report item when no licenses are in use?
    """
    get_output_from_server_mock.return_value = flexlm_output_no_licenses

    assert await flexlm_server.get_report_item(1, "testproduct.testfeature") == LicenseReportItem(
        feature_id=1,
        product_feature="testproduct.testfeature",
        used=0,
        total=1000,
        uses=[],
    )


@mark.asyncio
@mock.patch("lm_agent.server_interfaces.flexlm.run_command")
@mock.patch("lm_agent.server_interfaces.flexlm.FlexLMLicenseServer.get_commands_list")
async def test_flexlm_get_report_item_continues_on_exception(
    get_commands_list_mock: mock.MagicMock,
    run_command_mock: mock.MagicMock,
    flexlm_server: FlexLMLicenseServer,
    flexlm_output: str,
):
    """
    Do the FlexLM server interface check the next command if the previous one fails?
    """
    get_commands_list_mock.return_value = [
        [
            f"{settings.LMUTIL_PATH}",
            "lmstat",
            "-c",
            "2345@127.0.0.1",
            "-f",
        ],
        [
            f"{settings.LMUTIL_PATH}",
            "lmstat",
            "-c",
            "3456@127.0.0.1",
            "-f",
        ],
    ]
    run_command_mock.side_effect = [CommandFailedToExecute("Command failed for first server"), flexlm_output]

    assert await flexlm_server.get_report_item(1, "testproduct.testfeature") == LicenseReportItem(
        feature_id=1,
        product_feature="testproduct.testfeature",
        used=93,
        total=1000,
        uses=[
            {"username": "sdmfva", "lead_host": "myserver.example.com", "booked": 29},
            {"username": "adfdna", "lead_host": "myserver.example.com", "booked": 27},
            {"username": "sdmfva", "lead_host": "myserver.example.com", "booked": 37},
        ],
    )



from collections import defaultdict
from unittest import mock

from pytest import mark

from lm_agent.models import (
    BookingSchema,
    ExtractedBookingSchema,
    ExtractedUsageSchema,
    JobSchema,
    LicenseReportItem,
)
from lm_agent.services.clean_jobs_and_bookings import (
    clean_bookings_by_usage,
    clean_jobs_and_bookings,
    clean_jobs_by_grace_time,
    clean_jobs_no_longer_running,
    clean_jobs_without_bookings,
    extract_bookings_from_job,
    extract_usages_from_report,
    get_bookings_mapping,
    get_cluster_grace_times,
    get_greatest_grace_time_for_job,
    get_usages_mapping,
)


def test_get_cluster_grace_times(parsed_configurations):
    """
    Test that get_cluster_grace_times generates a dict with the grace time for each feature_id.
    """
    expected_grace_times = {1: 60, 2: 123}

    result = get_cluster_grace_times(parsed_configurations)
    assert result == expected_grace_times


def test__get_greatest_grace_time_for_job(parsed_jobs):
    """
    Test if the function really returns the greatest value for the grace_time.
    """
    grace_times = {
        1: 10,
        2: 20,
    }
    job_bookings = parsed_jobs[0].bookings
    expected_result = 20

    result = get_greatest_grace_time_for_job(grace_times, job_bookings)
    assert result == expected_result


@mark.parametrize(
    "parsed_job,extracted_bookings",
    [
        (
            "one_parsed_job",
            [
                ExtractedBookingSchema(
                    booking_id=1,
                    job_id=1,
                    slurm_job_id="123",
                    username="user1",
                    lead_host="host1",
                    feature_id=1,
                    quantity=12,
                ),
                ExtractedBookingSchema(
                    booking_id=2,
                    job_id=1,
                    slurm_job_id="123",
                    username="user1",
                    lead_host="host1",
                    feature_id=2,
                    quantity=50,
                ),
            ],
        ),
        (
            "another_parsed_job",
            [
                ExtractedBookingSchema(
                    booking_id=3,
                    job_id=2,
                    slurm_job_id="456",
                    username="user2",
                    lead_host="host2",
                    feature_id=4,
                    quantity=15,
                ),
                ExtractedBookingSchema(
                    booking_id=4,
                    job_id=2,
                    slurm_job_id="456",
                    username="user2",
                    lead_host="host2",
                    feature_id=7,
                    quantity=25,
                ),
            ],
        ),
    ],
)
def test__extract_bookings_from_job(parsed_job, extracted_bookings, request):
    """
    Test that the bookings can be extrated from a job.
    """
    parsed_job = request.getfixturevalue(parsed_job)

    assert extract_bookings_from_job(parsed_job) == extracted_bookings


@mark.parametrize(
    "report_item,extracted_usages",
    [
        (
            "one_report_item",
            [
                ExtractedUsageSchema(feature_id=1, username="user1", lead_host="host1", quantity=100),
            ],
        ),
        (
            "another_report_item",
            [
                ExtractedUsageSchema(feature_id=2, username="user2", lead_host="host2", quantity=10),
            ],
        ),
    ],
)
def test__extract_usages_from_report(report_item, extracted_usages, request):
    """
    Test that the usages can be extracted from a report item.
    """
    report_item = request.getfixturevalue(report_item)

    assert extract_usages_from_report(report_item) == extracted_usages


def test_get_bookings_mapping(parsed_jobs):
    """
    Test tha the bookings can be mapped by the required information.
    """
    result = defaultdict(
        list,
        {
            (1, "user1", "host1", 100): [
                ExtractedBookingSchema(
                    booking_id=1,
                    job_id=1,
                    slurm_job_id="123",
                    username="user1",
                    lead_host="host1",
                    feature_id=1,
                    quantity=100,
                )
            ],
            (2, "user1", "host1", 50): [
                ExtractedBookingSchema(
                    booking_id=2,
                    job_id=1,
                    slurm_job_id="123",
                    username="user1",
                    lead_host="host1",
                    feature_id=2,
                    quantity=50,
                )
            ],
            (1, "user2", "host2", 15): [
                ExtractedBookingSchema(
                    booking_id=3,
                    job_id=2,
                    slurm_job_id="456",
                    username="user2",
                    lead_host="host2",
                    feature_id=1,
                    quantity=15,
                )
            ],
            (2, "user2", "host2", 25): [
                ExtractedBookingSchema(
                    booking_id=4,
                    job_id=2,
                    slurm_job_id="456",
                    username="user2",
                    lead_host="host2",
                    feature_id=2,
                    quantity=25,
                )
            ],
            (1, "user3", "host3", 5): [
                ExtractedBookingSchema(
                    booking_id=14,
                    job_id=3,
                    slurm_job_id="789",
                    username="user3",
                    lead_host="host3",
                    feature_id=1,
                    quantity=5,
                )
            ],
            (2, "user3", "host3", 17): [
                ExtractedBookingSchema(
                    booking_id=15,
                    job_id=3,
                    slurm_job_id="789",
                    username="user3",
                    lead_host="host3",
                    feature_id=2,
                    quantity=17,
                )
            ],
        },
    )

    assert get_bookings_mapping(parsed_jobs) == result


def test__get_usages_mapping(parsed_report_items):
    """
    Test that the usages can be mapped by the required information.
    """
    result = defaultdict(
        list,
        {
            (1, "user1", "host1", 100): [
                ExtractedUsageSchema(feature_id=1, username="user1", lead_host="host1", quantity=100)
            ],
            (2, "user2", "host2", 10): [
                ExtractedUsageSchema(feature_id=2, username="user2", lead_host="host2", quantity=10)
            ],
        },
    )

    assert get_usages_mapping(parsed_report_items) == result


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_job_by_slurm_job_id")
async def test__clean_jobs_without_bookings__removes_job(
    remove_job_mock, one_parsed_job, parsed_job_without_bookings
):
    """
    Test that the function removes the job when there are no bookings in it.
    """
    cluster_jobs = [one_parsed_job, parsed_job_without_bookings]

    remaining_jobs_with_bookings = await clean_jobs_without_bookings(cluster_jobs)

    remove_job_mock.assert_called_once_with(parsed_job_without_bookings.slurm_job_id)
    assert remaining_jobs_with_bookings == [one_parsed_job]


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_job_by_slurm_job_id")
async def test__clean_jobs_without_bookings__no_jobs_to_remove(remove_job_mock, one_parsed_job):
    """
    Test that the function doesn't remove any job when all jobs have bookings.
    """
    cluster_jobs = [one_parsed_job]

    remaining_jobs_with_bookings = await clean_jobs_without_bookings(cluster_jobs)

    remove_job_mock.assert_not_called()
    assert remaining_jobs_with_bookings == cluster_jobs


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_job_by_slurm_job_id")
async def test__clean_jobs_no_longer_running__not_running(remove_job_by_slurm_job_id_mock, parsed_jobs):
    """
    Check that the jobs that aren't running are cleaned.
    """
    squeue_result = [
        {"job_id": 123, "state": "RUNNING"},
        {"job_id": 456, "state": "COMPLETED"},
        {"job_id": 789, "state": "SUSPENDED"},
    ]

    remaining_jobs_with_bookings = await clean_jobs_no_longer_running(parsed_jobs, squeue_result)

    assert remove_job_by_slurm_job_id_mock.call_count == 2
    remove_job_by_slurm_job_id_mock.assert_has_calls([mock.call("456"), mock.call("789")])
    assert remaining_jobs_with_bookings == [parsed_jobs[0]]


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_job_by_slurm_job_id")
async def test__clean_jobs_no_longer_running__not_in_squeue(remove_job_by_slurm_job_id_mock, parsed_jobs):
    """
    Check that the jobs that aren't in the squeue result are cleaned.
    """
    squeue_result = [
        {"job_id": 123, "state": "RUNNING"},
    ]

    remaining_jobs_with_bookings = await clean_jobs_no_longer_running(parsed_jobs, squeue_result)

    assert remove_job_by_slurm_job_id_mock.call_count == 2
    remove_job_by_slurm_job_id_mock.assert_has_calls([mock.call("456"), mock.call("789")])
    assert remaining_jobs_with_bookings == [parsed_jobs[0]]


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_job_by_slurm_job_id")
async def test__clean_jobs_no_longer_running__no_squeue_result(remove_job_by_slurm_job_id_mock, parsed_jobs):
    """
    Check that all jobs are cleaned if there are no jobs in the squeue result.
    """
    squeue_result = []

    remaining_jobs_with_bookings = await clean_jobs_no_longer_running(parsed_jobs, squeue_result)

    assert remove_job_by_slurm_job_id_mock.call_count == 3
    remove_job_by_slurm_job_id_mock.assert_has_calls([mock.call("123"), mock.call("456"), mock.call("789")])
    assert remaining_jobs_with_bookings == []


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_job_by_slurm_job_id")
async def test__clean_jobs_by_grace_time__grace_time_expired(
    remove_job_by_slurm_job_id_mock,
    parsed_jobs,
):
    """
    Check that the job is cleaned when the running time is greater than grace_time.
    """
    slurm_job_id = "123"

    squeue_result = [
        {
            "job_id": 123,
            "run_time_in_seconds": 300,
            "state": "RUNNING",
        }
    ]
    grace_times = {1: 10, 2: 20, 4: 10, 7: 20}

    remaining_jobs = await clean_jobs_by_grace_time(parsed_jobs, squeue_result, grace_times)

    remove_job_by_slurm_job_id_mock.assert_awaited_once_with(slurm_job_id)
    assert len(remaining_jobs) == 2


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_job_by_slurm_job_id")
async def test__clean_jobs_by_grace_time__within_grace_time(
    remove_job_by_slurm_job_id_mock,
    parsed_jobs,
):
    """
    Check that the job is not cleaned when the running time is within the grace_time.
    """
    squeue_result = [
        {
            "job_id": 123,
            "run_time_in_seconds": 300,
            "state": "RUNNING",
        }
    ]
    grace_times = {1: 1000, 2: 3000, 4: 500, 7: 1500}

    remaining_jobs = await clean_jobs_by_grace_time(parsed_jobs, squeue_result, grace_times)

    remove_job_by_slurm_job_id_mock.assert_not_called()
    assert remaining_jobs == parsed_jobs


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_job_by_slurm_job_id")
async def test__clean_jobs_by_grace_time__no_jobs(
    remove_job_by_slurm_job_id_mock,
):
    """
    Check that the function doesn't do anything when there are no jobs.
    """
    squeue_result = []
    grace_times = {1: 1000, 2: 3000}

    remaining_jobs = await clean_jobs_by_grace_time([], squeue_result, grace_times)

    remove_job_by_slurm_job_id_mock.assert_not_called()
    assert remaining_jobs == []


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_booking")
async def test__clean_bookings_by_usage__only_one_booking(remove_bookings_mock):
    """
    Test that the bookings can be cleaned by usage when there is only one booking.
    """
    cluster_jobs = [
        JobSchema(
            id=1,
            slurm_job_id="123",
            cluster_client_id="dummy",
            username="user1",
            lead_host="host1",
            bookings=[
                BookingSchema(id=1, job_id=1, feature_id=1, quantity=12),
            ],
        ),
    ]

    report_items = [
        LicenseReportItem(
            feature_id=1,
            product_feature="product.feature",
            used=24,
            total=100,
            uses=[
                {"username": "user1", "lead_host": "host1", "booked": 12},
                {"username": "user2", "lead_host": "host2", "booked": 12},
            ],
        )
    ]

    await clean_bookings_by_usage(cluster_jobs, report_items)

    remove_bookings_mock.assert_called_once_with(1)


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_booking")
async def test__clean_bookings_by_usage__multiple_bookings(remove_bookings_mock):
    cluster_jobs = [
        JobSchema(
            id=1,
            slurm_job_id="123",
            cluster_client_id="dummy",
            username="user1",
            lead_host="host1",
            bookings=[
                BookingSchema(id=1, job_id=1, feature_id=1, quantity=12),
                BookingSchema(id=2, job_id=1, feature_id=2, quantity=50),
            ],
        ),
    ]

    report_items = [
        LicenseReportItem(
            feature_id=1,
            product_feature="product1.feature1",
            used=24,
            total=100,
            uses=[
                {"username": "user1", "lead_host": "host1", "booked": 12},
            ],
        ),
        LicenseReportItem(
            feature_id=2,
            product_feature="product2.feature2",
            used=50,
            total=100,
            uses=[
                {"username": "user1", "lead_host": "host1", "booked": 50},
            ],
        ),
    ]

    await clean_bookings_by_usage(cluster_jobs, report_items)

    remove_bookings_mock.assert_has_calls([mock.call(1), mock.call(2)])


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_booking")
async def test__clean_bookings_by_usage__multiple_jobs(remove_bookings_mock):
    cluster_jobs = [
        JobSchema(
            id=1,
            slurm_job_id="123",
            cluster_client_id="dummy",
            username="user1",
            lead_host="host1",
            bookings=[
                BookingSchema(id=1, job_id=1, feature_id=1, quantity=12),
            ],
        ),
        JobSchema(
            id=2,
            slurm_job_id="456",
            cluster_client_id="dummy",
            username="user2",
            lead_host="host2",
            bookings=[
                BookingSchema(id=2, job_id=2, feature_id=1, quantity=50),
            ],
        ),
    ]

    report_items = [
        LicenseReportItem(
            feature_id=1,
            product_feature="product.feature",
            used=24,
            total=100,
            uses=[
                {"username": "user1", "lead_host": "host1", "booked": 12},
                {"username": "user2", "lead_host": "host2", "booked": 50},
            ],
        ),
    ]

    await clean_bookings_by_usage(cluster_jobs, report_items)

    remove_bookings_mock.assert_has_calls([mock.call(1), mock.call(2)])


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_booking")
async def test__clean_bookings_by_usage__multiple_bookings_with_same_information(remove_bookings_mock):
    cluster_jobs = [
        JobSchema(
            id=1,
            slurm_job_id="123",
            cluster_client_id="dummy",
            username="user1",
            lead_host="host1",
            bookings=[
                BookingSchema(id=1, job_id=1, feature_id=1, quantity=12),
            ],
        ),
        JobSchema(
            id=2,
            slurm_job_id="234",
            cluster_client_id="dummy",
            username="user1",
            lead_host="host1",
            bookings=[
                BookingSchema(id=2, job_id=2, feature_id=1, quantity=12),
            ],
        ),
    ]

    report_items = [
        LicenseReportItem(
            feature_id=1,
            product_feature="product.feature",
            used=24,
            total=100,
            uses=[
                {"username": "user1", "lead_host": "host1", "booked": 12},
                {"username": "user1", "lead_host": "host1", "booked": 12},
            ],
        ),
    ]

    await clean_bookings_by_usage(cluster_jobs, report_items)

    remove_bookings_mock.assert_has_calls([mock.call(1), mock.call(2)])


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_booking")
async def test__clean_bookings_by_usage__more_bookings_than_usage(remove_booking_mock):
    cluster_jobs = [
        JobSchema(
            id=1,
            slurm_job_id="123",
            cluster_client_id="dummy",
            username="user1",
            lead_host="host1",
            bookings=[
                BookingSchema(id=1, job_id=1, feature_id=1, quantity=12),
            ],
        ),
        JobSchema(
            id=2,
            slurm_job_id="234",
            cluster_client_id="dummy",
            username="user1",
            lead_host="host1",
            bookings=[
                BookingSchema(id=2, job_id=2, feature_id=1, quantity=12),
            ],
        ),
    ]

    report_items = [
        LicenseReportItem(
            feature_id=1,
            product_feature="product.feature",
            used=12,
            total=100,
            uses=[
                {"username": "user1", "lead_host": "host1", "booked": 12},
            ],
        ),
    ]

    await clean_bookings_by_usage(cluster_jobs, report_items)

    remove_booking_mock.assert_not_called()


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_booking")
async def test__clean_bookings_by_usage__more_usages_than_bookings(remove_booking_mock):
    cluster_jobs = [
        JobSchema(
            id=1,
            slurm_job_id="123",
            cluster_client_id="dummy",
            username="user1",
            lead_host="host1",
            bookings=[
                BookingSchema(id=1, job_id=1, feature_id=1, quantity=12),
            ],
        ),
    ]

    report_items = [
        LicenseReportItem(
            feature_id=1,
            product_feature="product.feature",
            used=24,
            total=100,
            uses=[
                {"username": "user1", "lead_host": "host1", "booked": 12},
                {"username": "user1", "lead_host": "host1", "booked": 12},
            ],
        ),
    ]

    await clean_bookings_by_usage(cluster_jobs, report_items)

    remove_booking_mock.assert_not_called()


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_booking")
async def test__clean_bookings_by_usage__no_bookings(remove_booking_mock):
    cluster_jobs = [
        JobSchema(
            id=1,
            slurm_job_id="123",
            cluster_client_id="dummy",
            username="user1",
            lead_host="host1",
            bookings=[],
        ),
    ]

    report_items = [
        LicenseReportItem(
            feature_id=1,
            product_feature="product1.feature1",
            used=24,
            total=100,
            uses=[
                {"username": "user1", "lead_host": "host1", "booked": 12},
            ],
        ),
        LicenseReportItem(
            feature_id=2,
            product_feature="product2.feature2",
            used=50,
            total=100,
            uses=[
                {"username": "user1", "lead_host": "host1", "booked": 50},
            ],
        ),
    ]

    await clean_bookings_by_usage(cluster_jobs, report_items)

    remove_booking_mock.assert_not_called()


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_job_by_slurm_job_id")
async def test__clean_jobs_and_bookings__remove_jobs_without_bookings(
    remove_job_by_slurm_job_id_mock,
    parsed_configurations,
    parsed_report_items,
    one_parsed_job,
    parsed_job_without_bookings,
):
    cluster_jobs = [
        one_parsed_job,
        parsed_job_without_bookings,
    ]

    squeue_result = [
        {"job_id": 123, "run_time_in_seconds": 15, "state": "RUNNING"},
        {"job_id": 789, "run_time_in_seconds": 30, "state": "RUNNING"},
    ]

    await clean_jobs_and_bookings(parsed_configurations, cluster_jobs, squeue_result, parsed_report_items)

    remove_job_by_slurm_job_id_mock.assert_called_once_with("789")


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_job_by_slurm_job_id")
async def test__clean_jobs_and_bookings__remove_jobs_no_longer_running(
    remove_job_by_slurm_job_id_mock, parsed_configurations, parsed_report_items, parsed_jobs
):
    squeue_result = [
        {"job_id": 456, "run_time_in_seconds": 15, "state": "RUNNING"},
        {"job_id": 789, "run_time_in_seconds": 30, "state": "RUNNING"},
    ]

    await clean_jobs_and_bookings(parsed_configurations, parsed_jobs, squeue_result, parsed_report_items)

    remove_job_by_slurm_job_id_mock.assert_called_once_with("123")


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_job_by_slurm_job_id")
async def test__clean_jobs_and_bookings__clean_jobs_by_grace_time(
    remove_job_by_slurm_job_id_mock, parsed_configurations, parsed_report_items, parsed_jobs
):
    squeue_result = [
        {"job_id": 123, "run_time_in_seconds": 300, "state": "RUNNING"},
        {"job_id": 456, "run_time_in_seconds": 30, "state": "RUNNING"},
        {"job_id": 789, "run_time_in_seconds": 30, "state": "RUNNING"},
    ]

    await clean_jobs_and_bookings(parsed_configurations, parsed_jobs, squeue_result, parsed_report_items)

    remove_job_by_slurm_job_id_mock.assert_called_once_with("123")


@mark.asyncio
@mock.patch("lm_agent.services.clean_jobs_and_bookings.remove_booking")
async def test__clean_jobs_and_bookings__clean_bookings_by_usage(
    remove_booking_mock, parsed_configurations, parsed_report_items, parsed_jobs
):
    squeue_result = [
        {"job_id": 123, "run_time_in_seconds": 15, "state": "RUNNING"},
        {"job_id": 456, "run_time_in_seconds": 30, "state": "RUNNING"},
        {"job_id": 789, "run_time_in_seconds": 30, "state": "RUNNING"},
    ]

    await clean_jobs_and_bookings(parsed_configurations, parsed_jobs, squeue_result, parsed_report_items)

    remove_booking_mock.assert_called_once_with(1)



"""
Test the environment parser the creates Settings.
"""

import os
from unittest.mock import patch

from pytest import fixture, raises

from lm_agent.config import init_settings


@fixture
def good_env():
    """
    A parseable environment
    """
    env = {"LM_AGENT_BACKEND_BASE_URL": "http://hello/"}
    with patch.dict(os.environ, env) as e:
        yield e


@fixture
def bad_env():
    """
    An unparseable environment
    """
    env = {"LM_AGENT_BACKEND_BASE_URL": "not-a-url"}
    with patch.dict(os.environ, env) as e:
        yield e


def test_init_settings(good_env):
    """
    Do we build a settings object from good input?
    """
    good = init_settings()
    assert str(good.BACKEND_BASE_URL) == good_env["LM_AGENT_BACKEND_BASE_URL"]


def test_init_settings_bad(bad_env, caplog):
    """
    Do we bail out when bad settings?
    """
    with raises(SystemExit):
        init_settings()
    assert (
        len(caplog.messages) == 1
        and "Input should be a valid URL, relative URL without a base [type=url_parsing, input_value='not-a-url', input_type=str]"  # noqa: E501
        in caplog.text
    )



"""
Test Epilog script.
"""

from unittest import mock

import pytest

from lm_agent.workload_managers.slurm.slurmctld_epilog import epilog


@pytest.mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_epilog.get_job_context")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_epilog.reconcile")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_epilog.get_required_licenses_for_job")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_epilog.remove_job_by_slurm_job_id")
async def test_epilog(
    remove_job_by_slurm_job_id_mock,
    get_required_licenses_for_job_mock,
    reconcile_mock,
    get_job_context_mock,
):
    bookings_mock = mock.MagicMock()
    bookings_mock.product_feature = "test.feature"
    bookings_mock.quantity = 10
    get_required_licenses_for_job_mock.return_value = [bookings_mock]

    get_job_context_mock.return_value = {
        "job_id": "1",
        "user_name": "user1",
        "lead_host": "host1",
        "cluster_name": "cluster1",
        "job_licenses": "test.feature@flexlm:10",
    }
    await epilog()
    reconcile_mock.assert_awaited_once()
    get_required_licenses_for_job_mock.assert_called_once_with("test.feature@flexlm:10")
    remove_job_by_slurm_job_id_mock.assert_awaited_once_with("1")


@pytest.mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_epilog.settings")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_epilog.get_job_context")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_epilog.reconcile")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_epilog.get_required_licenses_for_job")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_epilog.remove_job_by_slurm_job_id")
async def test_epilog_without_triggering_reconcile(
    remove_job_by_slurm_job_id_mock,
    get_required_licenses_for_job_mock,
    reconcile_mock,
    get_job_context_mock,
    settings_mock,
):
    bookings_mock = mock.MagicMock()
    bookings_mock.product_feature = "test.feature"
    bookings_mock.quantity = 10
    get_required_licenses_for_job_mock.return_value = [bookings_mock]

    get_job_context_mock.return_value = {
        "job_id": "1",
        "user_name": "user1",
        "lead_host": "host1",
        "cluster_name": "cluster1",
        "job_licenses": "test.feature@flexlm:10",
    }

    settings_mock.USE_RECONCILE_IN_PROLOG_EPILOG = False

    await epilog()
    reconcile_mock.assert_not_called()
    get_required_licenses_for_job_mock.assert_called_once_with("test.feature@flexlm:10")
    remove_job_by_slurm_job_id_mock.assert_awaited_once_with("1")



from unittest import mock

from pytest import mark

from lm_agent.models import LicenseReportItem
from lm_agent.services.reconciliation import reconcile


@mark.asyncio
@mock.patch("lm_agent.services.reconciliation.return_formatted_squeue_out")
@mock.patch("lm_agent.services.reconciliation.create_or_update_reservation")
@mock.patch("lm_agent.services.reconciliation.get_all_features_cluster_values")
@mock.patch("lm_agent.services.reconciliation.get_cluster_configs_from_backend")
@mock.patch("lm_agent.services.reconciliation.get_all_features_bookings_sum")
@mock.patch("lm_agent.services.reconciliation.get_cluster_jobs_from_backend")
@mock.patch("lm_agent.services.reconciliation.update_features")
async def test__reconcile__success(
    update_features_mock,
    get_jobs_from_backend_mock,
    get_bookings_sum_mock,
    get_configs_from_backend_mock,
    get_all_cluster_values_mock,
    create_or_update_reservation_mock,
    return_formatted_squeue_out_mock,
    parsed_configurations,
):
    """
    Check if reconcile updates the reservation with the correct value.
    The reservation should block all licenses that are in use.

    License: abaqus.abaqus@flexlm
    Total: 1000

    Cluster: cluster1
    Used in cluster: 23

    Overview of the license:
    _________________________________________________________________
    |    200   |    15     |     17    |    71    |       697       |
    |   used   |   booked  |   booked  |  booked  |      free       |
    | Lic serv | cluster 1 | cluster 2 | cluster3 |     to use      |
    -----------------------------------------------------------------

    Since we have 303 licenses in use (booked or license server), the
    amount of licenses available is 697.

    This way, we need to block the 303 licenses that are in use.
    But considering that Slurm is already "blocking" 23 licenses
    that are in use in the cluster, the reservation should block 280 licenses.

    Formula to calculate the reservation:
    reservation = lic serv used - slurm used + booked sum
    reservation = 200 - 23 + 103 = 280

    """
    update_features_mock.return_value = [
        LicenseReportItem(
            feature_id=1,
            product_feature="abaqus.abaqus",
            total=1000,
            used=200,
            uses=[],
        )
    ]
    get_configs_from_backend_mock.return_value = parsed_configurations
    get_jobs_from_backend_mock.return_value = []
    get_bookings_sum_mock.return_value = {"abaqus.abaqus": 103}
    get_all_cluster_values_mock.return_value = {"abaqus.abaqus": {"total": 1000, "used": 23}}
    return_formatted_squeue_out_mock.return_value = ""

    await reconcile()
    create_or_update_reservation_mock.assert_called_with("abaqus.abaqus@flexlm:280")



"""
Test the LM-X parser.
"""

from pytest import mark

from lm_agent.models import LicenseUsesItem
from lm_agent.parsing.lmx import parse, parse_feature_line, parse_in_use_line, parse_usage_line


def test_parse_feature_line():
    """
    Does the regex for the feature line match the lines in the output?
    The line contains:
    - feature
    - version
    - vendor
    From these, we only need to extract ``feature``.
    """
    assert parse_feature_line("Feature: CatiaV5Reader Version: 21.0 Vendor: ALTAIR") == "catiav5reader"
    assert parse_feature_line("Feature: GlobalZoneEU Version: 21.0 Vendor: ALTAIR") == "globalzoneeu"
    assert parse_feature_line("Feature: HWAIFPBS Version: 21.0 Vendor: ALTAIR") == "hwaifpbs"
    assert parse_feature_line("not a feature line") is None
    assert parse_feature_line("Version: 21.0 Vendor: ALTAIR") is None
    assert parse_feature_line("") is None


def test_parse_in_use_line():
    """
    Does the regex for the in use line match the lines in the output?
    The line contains:
    - in_use
    - total
    """
    assert parse_in_use_line("0 of 3 license(s) used") == {
        "used": 0,
        "total": 3,
    }
    assert parse_in_use_line("288000 of 1000003 license(s) used:") == {
        "used": 288000,
        "total": 1000003,
    }
    assert parse_in_use_line("aaaaa of bbbbbbb licenses") is None
    assert parse_in_use_line("") is None


def test_parse_usage_line():
    """
    Does the regex for the usage line match the line in the output?
    The line contains:
    - user name
    - lead host
    - booked
    """
    assert parse_usage_line("15000 license(s) used by VRAAFG@RD0082879 [138.106.159.158]") == LicenseUsesItem(
        username="vraafg",
        lead_host="RD0082879",
        booked=15000,
    )
    assert parse_usage_line(
        "25000 license(s) used by sbak8y@p-c39.maas.rnd.com [10.104.193.54]"
    ) == LicenseUsesItem(
        username="sbak8y",
        lead_host="p-c39.maas.rnd.com",
        booked=25000,
    )
    assert parse_usage_line(
        "15000 license(s) used by sssegm@p-g2.maas.rnd.com [10.104.192.204]"
    ) == LicenseUsesItem(
        username="sssegm",
        lead_host="p-g2.maas.rnd.com",
        booked=15000,
    )
    assert parse_usage_line(
        "1 license(s) used by mbrzy5@dcv046.com_ver2023 [10.123.321.20]"
    ) == LicenseUsesItem(
        username="mbrzy5",
        lead_host="dcv046.com",
        booked=1,
    )
    assert parse_usage_line(
        "1 license(s) used by k12dca@ms0904_ver5.4.1 [10.123.321.156]"
    ) == LicenseUsesItem(
        username="k12dca",
        lead_host="ms0904",
        booked=1,
    )
    assert parse_usage_line("0 license(s) used by v-c54.aaa.aa") is None
    assert parse_usage_line("") is None


@mark.parametrize(
    "fixture,result",
    [
        (
            "lmx_output",
            {
                "catiav5reader": {"total": 3, "used": 0, "uses": []},
                "globalzoneeu": {
                    "total": 1000003,
                    "used": 40000,
                    "uses": [
                        LicenseUsesItem(username="vraafg", lead_host="RD0082879", booked=15000),
                        LicenseUsesItem(username="vraafg", lead_host="RD0082879", booked=25000),
                    ],
                },
                "hwaifpbs": {"total": 2147483647, "used": 0, "uses": []},
                "hwawpf": {"total": 2147483647, "used": 0, "uses": []},
                "hwactivate": {"total": 2147483647, "used": 0, "uses": []},
                "hwflux2d": {
                    "total": 2147483647,
                    "used": 30000,
                    "uses": [
                        LicenseUsesItem(username="vraafg", lead_host="RD0082879", booked=15000),
                        LicenseUsesItem(username="vraafg", lead_host="RD0082879", booked=15000),
                    ],
                },
                "hyperworks": {
                    "total": 1000000,
                    "used": 25000,
                    "uses": [
                        LicenseUsesItem(username="sssaah", lead_host="RD0082406", booked=25000),
                    ],
                },
            },
        ),
        (
            "lmx_output_2",
            {
                "femfat_visualizer": {
                    "total": 2,
                    "used": 2,
                    "uses": [
                        LicenseUsesItem(booked=1, lead_host="dcv046.com", username="fdsva1"),
                        LicenseUsesItem(booked=1, lead_host="dcv048.com", username="asdsc1"),
                    ],
                },
            },
        ),
    ],
)
def test_parse__correct_output(request, fixture, result):
    """
    Does the parser return the correct data for this output?
    - lmx_output: expected output from the license server,
    which contain licenses and usage information.
    """
    output = request.getfixturevalue(fixture)
    assert parse(output) == result


def test_parse__bad_output(lmx_output_bad):
    """
    Does the parser return the correct data for this output?
    - lmx_output_bad: unparseable output from the license server,
    which can happen when a connection error occours.
    """
    assert parse(lmx_output_bad) == {}


def test_parse__no_licenses_output(lmx_output_no_licenses):
    """
    Does the parser return the correct data for this output?
    - lmx_output_no_licenses: expected output from the server
    when none of the licenses are in use by users.
    """
    assert parse(lmx_output_no_licenses) == {
        "catiav5reader": {"total": 3, "used": 0, "uses": []},
        "globalzoneeu": {"total": 1000003, "used": 0, "uses": []},
        "hwaifpbs": {"total": 2147483647, "used": 0, "uses": []},
        "hwawpf": {"total": 2147483647, "used": 0, "uses": []},
        "hwactivate": {"total": 2147483647, "used": 0, "uses": []},
        "hwflux2d": {"total": 2147483647, "used": 0, "uses": []},
        "hyperworks": {"total": 1000000, "used": 0, "uses": []},
    }



"""
Test Prolog script.
"""

from unittest import mock

import pytest

from lm_agent.models import FeatureSchema, ProductSchema
from lm_agent.workload_managers.slurm.slurmctld_prolog import prolog as main


@pytest.mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.get_job_context")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.get_required_licenses_for_job")
async def test_main_error_in_get_required_licenses_for_job(
    get_required_licenses_for_job_mock,
    get_job_context_mock,
):
    get_job_context_mock.return_value = {
        "job_id": "1",
        "user_name": "user1",
        "lead_host": "host1",
        "job_licenses": "",
    }
    get_required_licenses_for_job_mock.side_effect = Exception

    with pytest.raises(SystemExit) as exc_info:
        await main()

    assert exc_info.value.code == 1


@pytest.mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.get_job_context")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.get_required_licenses_for_job")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.get_cluster_configs_from_backend")
async def test_main_error_in_get_config_from_backend(
    get_configs_from_backend_mock,
    get_required_licenses_for_job_mock,
    get_job_context_mock,
):
    get_job_context_mock.return_value = {
        "job_id": "1",
        "user_name": "user1",
        "lead_host": "host1",
        "job_licenses": "test.feature@flexlm:10",
    }
    bookings_mock = mock.MagicMock()
    bookings_mock.product_feature = "test.feature"
    bookings_mock.quantity = 10
    get_required_licenses_for_job_mock.return_value = [bookings_mock]
    get_configs_from_backend_mock.side_effect = Exception

    with pytest.raises(SystemExit) as exc_info:
        await main()

    assert exc_info.value.code == 1

    get_required_licenses_for_job_mock.assert_called_once_with("test.feature@flexlm:10")


@pytest.mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.get_job_context")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.get_required_licenses_for_job")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.get_cluster_configs_from_backend")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.reconcile")
async def test_main_error_in_reconcile(
    reconcile_mock,
    get_configs_from_backend_mock,
    get_required_licenses_for_job_mock,
    get_job_context_mock,
):
    get_job_context_mock.return_value = {
        "job_id": "1",
        "user_name": "user1",
        "lead_host": "host1",
        "job_licenses": "test.feature@flexlm:10",
    }
    bookings_mock = mock.MagicMock()
    bookings_mock.product_feature = "test.feature"
    bookings_mock.quantity = 10
    get_required_licenses_for_job_mock.return_value = [bookings_mock]

    backend_return_mock = mock.MagicMock()
    backend_return_mock.features = [
        FeatureSchema(
            id=1,
            name="feature",
            product=ProductSchema(id=1, name="test"),
            config_id=1,
            reserved=100,
            total=1000,
            used=93,
            booked_total=0,
        )
    ]
    get_configs_from_backend_mock.return_value = [backend_return_mock]
    reconcile_mock.side_effect = Exception

    with pytest.raises(SystemExit) as exc_info:
        await main()

    assert exc_info.value.code == 1

    get_required_licenses_for_job_mock.assert_called_once_with("test.feature@flexlm:10")
    get_configs_from_backend_mock.assert_awaited_once()


@pytest.mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.sys")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.get_job_context")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.get_required_licenses_for_job")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.get_cluster_configs_from_backend")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.reconcile")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.make_booking_request")
async def test_main(
    make_booking_request_mock,
    reconcile_mock,
    get_configs_from_backend_mock,
    get_required_licenses_for_job_mock,
    get_job_context_mock,
    sys_mock,
):
    get_job_context_mock.return_value = {
        "job_id": "1",
        "user_name": "user1",
        "lead_host": "host1",
        "cluster_name": "cluster1",
        "job_licenses": "test.feature@flexlm:10",
    }
    bookings_mock = mock.MagicMock()
    bookings_mock.product_feature = "test.feature"
    bookings_mock.quantity = 10
    get_required_licenses_for_job_mock.return_value = [bookings_mock]

    backend_return_mock = mock.MagicMock()
    backend_return_mock.features = [
        FeatureSchema(
            id=1,
            name="feature",
            product=ProductSchema(id=1, name="test"),
            config_id=1,
            reserved=100,
            total=1000,
            used=93,
            booked_total=0,
        )
    ]
    get_configs_from_backend_mock.return_value = [backend_return_mock]

    await main()

    get_configs_from_backend_mock.assert_awaited_once()
    get_required_licenses_for_job_mock.assert_called_once_with("test.feature@flexlm:10")
    reconcile_mock.assert_awaited_once()
    make_booking_request_mock.assert_awaited_once()


@pytest.mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.sys")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.settings")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.get_job_context")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.get_required_licenses_for_job")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.get_cluster_configs_from_backend")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.reconcile")
@mock.patch("lm_agent.workload_managers.slurm.slurmctld_prolog.make_booking_request")
async def test_main_without_triggering_reconciliation(
    make_booking_request_mock,
    reconcile_mock,
    get_configs_from_backend_mock,
    get_required_licenses_for_job_mock,
    get_job_context_mock,
    settings_mock,
    sys_mock,
):
    get_job_context_mock.return_value = {
        "job_id": "1",
        "user_name": "user1",
        "lead_host": "host1",
        "cluster_name": "cluster1",
        "job_licenses": "test.feature@flexlm:10",
    }

    bookings_mock = mock.MagicMock()
    bookings_mock.product_feature = "test.feature"
    bookings_mock.quantity = 10
    get_required_licenses_for_job_mock.return_value = [bookings_mock]

    backend_return_mock = mock.MagicMock()
    backend_return_mock.features = [
        FeatureSchema(
            id=1,
            name="feature",
            product=ProductSchema(id=1, name="test"),
            config_id=1,
            reserved=100,
            total=1000,
            used=93,
            booked_total=0,
        )
    ]
    get_configs_from_backend_mock.return_value = [backend_return_mock]

    settings_mock.USE_RECONCILE_IN_PROLOG_EPILOG = False

    await main()

    get_configs_from_backend_mock.assert_awaited_once()
    get_required_licenses_for_job_mock.assert_called_once_with("test.feature@flexlm:10")
    make_booking_request_mock.assert_awaited_once()
    reconcile_mock.assert_not_called()



"""
Test the RLM parser.
"""

from pytest import mark

from lm_agent.models import LicenseUsesItem
from lm_agent.parsing.rlm import parse, parse_count_line, parse_feature_line, parse_usage_line


def test_parse_feature_line():
    """
    Does the regex for the feature line match the lines in the output?
    The line contains:
    - feature
    - version

    From these, we only need to extract ``feature``.
    """
    assert parse_feature_line("powercase v1.0") == "powercase"
    assert parse_feature_line("powerflow-decomp v1.0") == "powerflow-decomp"
    assert parse_feature_line("powerdelta-translate2 v1.0") == "powerdelta-translate2"
    assert parse_feature_line("not a feature line") is None
    assert parse_feature_line("Copyright (C) 2006-2017, Reprise Software, Inc. All rights reserved.") is None
    assert parse_feature_line("") is None


def test_parse_count_line():
    """
    Does the regex for the count line match the lines in the output?
    The line contains:
    - count
    - in_use
    """
    assert parse_count_line("count: 5, # reservations: 0, inuse: 3, exp: 31-dec-2023") == {
        "count": 5,
        "in_use": 3,
    }
    assert parse_count_line("count: 3000, # reservations: 0, inuse: 0, exp: 31-dec-2023") == {
        "count": 3000,
        "in_use": 0,
    }
    assert parse_count_line("count: 1000000, # reservations: 0, inuse: 0, exp: 31-dec-2023") == {
        "count": 1000000,
        "in_use": 0,
    }
    assert parse_count_line("aaaaa of bbbbbbb licenses") is None
    assert parse_count_line("") is None


def test_parse_usage_line():
    """
    Does the regex for the usage line match the line in the output?
    The line contains:
    - feature
    - user name
    - lead host
    - booked
    """
    assert parse_usage_line(
        "converge_super v3.0: asdj13@myserver.example.com 29/0 at 11/01 09:01  (handle: 15a)"
    ) == {
        "license_feature": "converge_super",
        "use": LicenseUsesItem(
            username="asdj13",
            lead_host="myserver.example.com",
            booked=29,
        ),
    }
    assert parse_usage_line("powercase v1.0: dfsdgv@server1 1/0 at 08/15 09:34  (handle: 182)") == {
        "license_feature": "powercase",
        "use": LicenseUsesItem(
            username="dfsdgv",
            lead_host="server1",
            booked=1,
        ),
    }
    assert parse_usage_line("aaaaa") is None
    assert parse_usage_line("") is None


@mark.parametrize(
    "fixture,result",
    [
        (
            "rlm_output",
            {
                "converge": {
                    "total": 1,
                    "used": 0,
                    "uses": [],
                },
                "converge_gui": {"total": 45, "used": 0, "uses": []},
                "converge_gui_polygonica": {"total": 1, "used": 0, "uses": []},
                "converge_super": {
                    "total": 1000,
                    "used": 93,
                    "uses": [
                        LicenseUsesItem(
                            booked=29,
                            lead_host="myserver.example.com",
                            username="asdj13",
                        ),
                        LicenseUsesItem(
                            booked=27,
                            lead_host="myserver.example.com",
                            username="cddcp2",
                        ),
                        LicenseUsesItem(
                            booked=37,
                            lead_host="myserver.example.com",
                            username="asdj13",
                        ),
                    ],
                },
                "converge_tecplot": {"total": 45, "used": 0, "uses": []},
            },
        ),
        (
            "another_rlm_output",
            {
                "powercase": {
                    "used": 3,
                    "total": 5,
                    "uses": [
                        LicenseUsesItem(username="dfsdgv", lead_host="server1", booked=1),
                        LicenseUsesItem(username="addvbh", lead_host="server2", booked=1),
                        LicenseUsesItem(username="wrtgb3", lead_host="server3", booked=1),
                    ],
                },
                "powerexport": {"used": 0, "total": 5, "uses": []},
                "powerflow-decomp": {"used": 0, "total": 3000, "uses": []},
                "powerflow-disc": {"used": 0, "total": 3000, "uses": []},
                "powerinsight": {"used": 0, "total": 5, "uses": []},
                "powerviz": {"used": 0, "total": 5, "uses": []},
                "powerviz-generator": {"used": 0, "total": 2, "uses": []},
                "powerviz-soiling": {"used": 0, "total": 7, "uses": []},
                "exasignalprocessingjob": {"used": 0, "total": 6, "uses": []},
                "poweracoustics": {"used": 0, "total": 1, "uses": []},
                "powercool": {"used": 0, "total": 1, "uses": []},
                "powerdelta": {
                    "used": 1,
                    "total": 1,
                    "uses": [LicenseUsesItem(username="ghnds2", lead_host="server4", booked=1)],
                },
                "powerdelta-meshunion": {"used": 0, "total": 1, "uses": []},
                "powerdelta-translate2": {"used": 0, "total": 1000000, "uses": []},
                "powerflow-sim": {"used": 0, "total": 7000, "uses": []},
                "powerinsight-generator": {"used": 0, "total": 1, "uses": []},
                "powerthermconnector": {"used": 0, "total": 1, "uses": []},
                "powertherm": {"used": 0, "total": 1, "uses": []},
                "powertherm-mp": {"used": 0, "total": 3, "uses": []},
                "powertherm-multigrid": {"used": 0, "total": 1, "uses": []},
                "powertherm-sc": {"used": 0, "total": 1, "uses": []},
                "powertherm-solver": {"used": 0, "total": 1, "uses": []},
            },
        ),
    ],
)
def test_parse__correct_ouput(request, fixture, result):
    """
    Does the parser return the correct data for this output?
    - rlm_output: expected output from the license server,
    which contain licenses and usage information.
    """
    output = request.getfixturevalue(fixture)
    assert parse(output) == result


def test_parse__bad_output(rlm_output_bad):
    """
    Does the parser return the correct data for this output?
    - rlm_output_bad: unparseable output from the license server,
    which can happen when a connection error occours.
    """
    assert parse(rlm_output_bad) == {}


def test_parse__no_licenses_output(rlm_output_no_licenses):
    """
    Does the parser return the correct data for this output?
    - rlm_output_no_licenses: expected output from the server
    when none of the licenses are in use by users.
    """
    assert parse(rlm_output_no_licenses) == {
        "converge": {"total": 1, "used": 0, "uses": []},
        "converge_gui": {"total": 45, "used": 0, "uses": []},
        "converge_gui_polygonica": {"total": 1, "used": 0, "uses": []},
        "converge_super": {"total": 1000, "used": 0, "uses": []},
        "converge_tecplot": {"total": 45, "used": 0, "uses": []},
    }



from unittest import mock

from httpx import Response
from pytest import mark, raises

from lm_agent.exceptions import LicenseManagerBackendConnectionError, LicenseManagerEmptyReportError
from lm_agent.models import (
    ConfigurationSchema,
    FeatureSchema,
    LicenseReportItem,
    LicenseServerSchema,
    LicenseServerType,
    ProductSchema,
)
from lm_agent.services import license_report


@mark.asyncio
@mark.parametrize(
    "output,reconciliation",
    [
        (
            "flexlm_output",
            [
                {
                    "feature_id": 1,
                    "product_feature": "testproduct.testfeature",
                    "used": 93,
                    "total": 1000,
                    "uses": [
                        {"booked": 29, "username": "sdmfva", "lead_host": "myserver.example.com"},
                        {"booked": 27, "username": "adfdna", "lead_host": "myserver.example.com"},
                        {"booked": 37, "username": "sdmfva", "lead_host": "myserver.example.com"},
                    ],
                },
            ],
        ),
    ],
)
@mock.patch("lm_agent.server_interfaces.flexlm.FlexLMLicenseServer.get_output_from_server")
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.scontrol_show_lic")
@mock.patch("lm_agent.services.license_report.get_cluster_configs_from_backend")
async def test_flexlm_get_report(
    get_configs_from_backend_mock: mock.MagicMock,
    show_lic_mock: mock.MagicMock,
    get_output_from_server_mock: mock.MagicMock,
    output,
    reconciliation,
    one_configuration_row_flexlm,
    scontrol_show_lic_output_flexlm,
    request,
):
    """
    Do I get a report for the FlexLM licenses in the cluster?
    """
    get_configs_from_backend_mock.return_value = [one_configuration_row_flexlm]
    show_lic_mock.return_value = scontrol_show_lic_output_flexlm

    output = request.getfixturevalue(output)
    get_output_from_server_mock.return_value = output

    reconcile_list = await license_report.report()
    assert reconcile_list == [LicenseReportItem(**item) for item in reconciliation]


@mark.asyncio
@mark.parametrize(
    "output,reconciliation",
    [
        (
            "rlm_output",
            [
                {
                    "feature_id": 1,
                    "product_feature": "converge.converge_super",
                    "used": 93,
                    "total": 1000,
                    "uses": [
                        {"booked": 29, "username": "asdj13", "lead_host": "myserver.example.com"},
                        {"booked": 27, "username": "cddcp2", "lead_host": "myserver.example.com"},
                        {"booked": 37, "username": "asdj13", "lead_host": "myserver.example.com"},
                    ],
                },
            ],
        ),
        (
            "rlm_output_no_licenses",
            [
                {
                    "feature_id": 1,
                    "product_feature": "converge.converge_super",
                    "used": 0,
                    "total": 1000,
                    "uses": [],
                },
            ],
        ),
    ],
)
@mock.patch("lm_agent.server_interfaces.rlm.RLMLicenseServer.get_output_from_server")
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.scontrol_show_lic")
@mock.patch("lm_agent.services.license_report.get_cluster_configs_from_backend")
async def test_rlm_get_report(
    get_configs_from_backend_mock: mock.MagicMock,
    show_lic_mock: mock.MagicMock,
    get_output_from_server_mock: mock.MagicMock,
    output,
    reconciliation,
    one_configuration_row_rlm,
    scontrol_show_lic_output_rlm,
    request,
):
    """
    Do I get a report for the RLM licenses in the cluster?
    """
    get_configs_from_backend_mock.return_value = [one_configuration_row_rlm]
    show_lic_mock.return_value = scontrol_show_lic_output_rlm

    output = request.getfixturevalue(output)
    get_output_from_server_mock.return_value = output

    reconcile_list = await license_report.report()
    assert reconcile_list == [LicenseReportItem(**item) for item in reconciliation]


@mark.asyncio
@mark.parametrize(
    "output,reconciliation",
    [
        (
            "lsdyna_output",
            [
                {
                    "feature_id": 1,
                    "product_feature": "mppdyna.mppdyna",
                    "used": 440,
                    "total": 500,
                    "uses": [
                        {"booked": 80, "username": "dvds3g", "lead_host": "n-c13.com"},
                        {"booked": 80, "username": "ssss1d", "lead_host": "n-c52.com"},
                        {"booked": 80, "username": "ssss1d", "lead_host": "n-c15.com"},
                        {"booked": 80, "username": "ywap0o", "lead_host": "n-c53.com"},
                        {"booked": 80, "username": "ywap0o", "lead_host": "n-c51.com"},
                        {"booked": 40, "username": "ndha1a", "lead_host": "n-c55.com"},
                    ],
                },
            ],
        ),
        (
            "lsdyna_output_no_licenses",
            [
                {
                    "feature_id": 1,
                    "product_feature": "mppdyna.mppdyna",
                    "used": 0,
                    "total": 500,
                    "uses": [],
                },
            ],
        ),
    ],
)
@mock.patch("lm_agent.server_interfaces.lsdyna.LSDynaLicenseServer.get_output_from_server")
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.scontrol_show_lic")
@mock.patch("lm_agent.services.license_report.get_cluster_configs_from_backend")
async def test_lsdyna_get_report(
    get_configs_from_backend_mock: mock.MagicMock,
    show_lic_mock: mock.MagicMock,
    get_output_from_server_mock: mock.MagicMock,
    output,
    reconciliation,
    one_configuration_row_lsdyna,
    scontrol_show_lic_output_lsdyna,
    request,
):
    """
    Do I get a report for the LS-Dyna licenses in the cluster?
    """
    get_configs_from_backend_mock.return_value = [one_configuration_row_lsdyna]
    show_lic_mock.return_value = scontrol_show_lic_output_lsdyna

    output = request.getfixturevalue(output)
    get_output_from_server_mock.return_value = output

    reconcile_list = await license_report.report()
    assert reconcile_list == [LicenseReportItem(**item) for item in reconciliation]


@mark.asyncio
@mark.parametrize(
    "output,reconciliation",
    [
        (
            "lmx_output",
            [
                LicenseReportItem(
                    feature_id=1,
                    product_feature="hyperworks.hyperworks",
                    used=25000,
                    total=1000000,
                    uses=[
                        {
                            "booked": 25000,
                            "username": "sssaah",
                            "lead_host": "RD0082406",
                        }
                    ],
                ),
            ],
        ),
        (
            "lmx_output_no_licenses",
            [
                LicenseReportItem(
                    feature_id=1,
                    product_feature="hyperworks.hyperworks",
                    used=0,
                    total=1000000,
                    uses=[],
                ),
            ],
        ),
    ],
)
@mock.patch("lm_agent.server_interfaces.lmx.LMXLicenseServer.get_output_from_server")
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.scontrol_show_lic")
@mock.patch("lm_agent.services.license_report.get_cluster_configs_from_backend")
async def test_lmx_get_report(
    get_configs_from_backend_mock: mock.MagicMock,
    show_lic_mock: mock.MagicMock,
    get_output_from_server_mock: mock.MagicMock,
    output,
    reconciliation,
    one_configuration_row_lmx,
    scontrol_show_lic_output_lmx: str,
    request,
):
    """
    Do I get a report for the LM-X licenses in the cluster?
    """
    get_configs_from_backend_mock.return_value = [one_configuration_row_lmx]
    show_lic_mock.return_value = scontrol_show_lic_output_lmx

    output = request.getfixturevalue(output)
    get_output_from_server_mock.return_value = output

    reconcile_list = await license_report.report()
    assert reconcile_list == reconciliation


@mark.asyncio
@mark.parametrize(
    "output,reconciliation",
    [
        (
            "olicense_output",
            [
                LicenseReportItem(
                    feature_id=1,
                    product_feature="cosin.ftire_adams",
                    used=3,
                    total=4,
                    uses=[
                        {"username": "sbhyma", "lead_host": "RD0087712", "booked": 1},
                        {"username": "sbhyma", "lead_host": "RD0087713", "booked": 1},
                        {"username": "user22", "lead_host": "RD0087713", "booked": 1},
                    ],
                ),
            ],
        ),
        (
            "olicense_output_no_licenses",
            [
                LicenseReportItem(
                    feature_id=1,
                    product_feature="cosin.ftire_adams",
                    used=0,
                    total=4,
                    uses=[],
                ),
            ],
        ),
    ],
)
@mock.patch("lm_agent.server_interfaces.olicense.OLicenseLicenseServer.get_output_from_server")
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.scontrol_show_lic")
@mock.patch("lm_agent.services.license_report.get_cluster_configs_from_backend")
async def test_olicense_get_report(
    get_configs_from_backend_mock: mock.MagicMock,
    show_lic_mock: mock.MagicMock,
    get_output_from_server_mock: mock.MagicMock,
    output,
    reconciliation,
    one_configuration_row_olicense,
    scontrol_show_lic_output_olicense: str,
    request,
):
    """
    Do I get a report for the OLicense licenses in the cluster?
    """
    get_configs_from_backend_mock.return_value = [one_configuration_row_olicense]
    show_lic_mock.return_value = scontrol_show_lic_output_olicense

    output = request.getfixturevalue(output)
    get_output_from_server_mock.return_value = output
    reconcile_list = await license_report.report()
    assert reconcile_list == reconciliation


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.scontrol_show_lic")
@mock.patch("lm_agent.services.license_report.get_cluster_configs_from_backend")
async def test_flexlm_report_with_empty_backend(
    get_configs_from_backend_mock: mock.MagicMock,
    show_lic_mock: mock.MagicMock,
    scontrol_show_lic_output_flexlm,
):
    """
    Do I collect the requested structured data when the backend is empty?
    """
    get_configs_from_backend_mock.return_value = []
    show_lic_mock.return_value = scontrol_show_lic_output_flexlm

    reconcile_list = await license_report.report()
    assert reconcile_list == []


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.scontrol_show_lic")
@mock.patch("lm_agent.services.license_report.get_cluster_configs_from_backend")
async def test_rlm_report_with_empty_backend(
    get_configs_from_backend_mock: mock.MagicMock,
    show_lic_mock: mock.MagicMock,
    scontrol_show_lic_output_rlm,
):
    """
    Do I collect the requested structured data when the backend is empty?
    """
    get_configs_from_backend_mock.return_value = []
    show_lic_mock.return_value = scontrol_show_lic_output_rlm

    reconcile_list = await license_report.report()
    assert reconcile_list == []


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.scontrol_show_lic")
@mock.patch("lm_agent.services.license_report.get_cluster_configs_from_backend")
async def test_lsdyna_report_with_empty_backend(
    get_configs_from_backend_mock: mock.MagicMock,
    show_lic_mock: mock.MagicMock,
    scontrol_show_lic_output_lsdyna,
):
    """
    Do I collect the requested structured data when the backend is empty?
    """
    get_configs_from_backend_mock.return_value = []
    show_lic_mock.return_value = scontrol_show_lic_output_lsdyna

    reconcile_list = await license_report.report()
    assert reconcile_list == []


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.scontrol_show_lic")
@mock.patch("lm_agent.services.license_report.get_cluster_configs_from_backend")
async def test_lmx_report_with_empty_backend(
    get_configs_from_backend_mock: mock.MagicMock,
    show_lic_mock: mock.MagicMock,
    scontrol_show_lic_output_lmx: str,
):
    """
    Do I collect the requested structured data when the backend is empty?
    """
    get_configs_from_backend_mock.return_value = []
    show_lic_mock.return_value = scontrol_show_lic_output_lmx

    reconcile_list = await license_report.report()
    assert reconcile_list == []


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.scontrol_show_lic")
@mock.patch("lm_agent.services.license_report.get_cluster_configs_from_backend")
async def test_olicense_report_with_empty_backend(
    get_configs_from_backend_mock: mock.MagicMock,
    show_lic_mock: mock.MagicMock,
    scontrol_show_lic_output_olicense: str,
):
    """
    Do I collect the requested structured data when the backend is empty?
    """
    get_configs_from_backend_mock.return_value = []
    show_lic_mock.return_value = scontrol_show_lic_output_olicense

    reconcile_list = await license_report.report()
    assert reconcile_list == []


def test_get_local_license_configurations():
    configuration_super = ConfigurationSchema(
        id=1,
        name="Converge",
        cluster_client_id="dummy",
        features=[
            FeatureSchema(
                id=1,
                name="converge_super",
                product=ProductSchema(id=1, name="converge"),
                config_id=1,
                reserved=100,
                total=1000,
                used=93,
                booked_total=0,
            )
        ],
        license_servers=[
            LicenseServerSchema(id=1, config_id=1, host="licserv0001", port=1234),
            LicenseServerSchema(id=3, config_id=1, host="licserv0003", port=8760),
        ],
        grace_time=60,
        type=LicenseServerType.RLM,
    )

    configuration_polygonica = ConfigurationSchema(
        id=2,
        name="Converge GUI Polygonica",
        cluster_client_id="dummy",
        features=[
            FeatureSchema(
                id=2,
                name="converge_gui_polygonica",
                product=ProductSchema(id=1, name="converge"),
                config_id=2,
                reserved=100,
                total=1000,
                used=93,
                booked_total=0,
            )
        ],
        license_servers=[
            LicenseServerSchema(id=1, config_id=2, host="licserv0001", port=1234),
            LicenseServerSchema(id=3, config_id=2, host="licserv0003", port=8760),
        ],
        grace_time=60,
        type=LicenseServerType.RLM,
    )

    license_configurations = [configuration_super, configuration_polygonica]
    local_licenses = ["converge.converge_super"]

    assert license_report.get_local_license_configurations(license_configurations, local_licenses) == [
        configuration_super
    ]


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.scontrol_show_lic")
@mock.patch("lm_agent.services.license_report.get_cluster_configs_from_backend")
@mock.patch("lm_agent.services.license_report.get_local_license_configurations")
@mock.patch("lm_agent.services.license_report.RLMLicenseServer.get_report_item")
async def test_license_report_empty_on_exception_raised(
    get_report_item_mock: mock.MagicMock,
    get_local_license_configurations_mock: mock.MagicMock,
    get_cluster_configs_from_backend_mock: mock.MagicMock,
    show_lic_mock: mock.MagicMock,
):
    """
    Do I get an empty report when an exception is raised?
    """
    get_cluster_configs_from_backend_mock.return_value = []
    show_lic_mock.return_value = ""
    get_local_license_configurations_mock.return_value = [
        ConfigurationSchema(
            id=1,
            name="Converge",
            cluster_client_id="dummy",
            features=[
                FeatureSchema(
                    id=1,
                    name="converge_super",
                    product=ProductSchema(id=1, name="converge"),
                    config_id=1,
                    reserved=100,
                    total=1000,
                    used=93,
                    booked_total=0,
                )
            ],
            license_servers=[
                LicenseServerSchema(id=1, config_id=1, host="licserv0001", port=1234),
                LicenseServerSchema(id=3, config_id=1, host="licserv0003", port=8760),
            ],
            grace_time=60,
            type=LicenseServerType.RLM,
        )
    ]
    get_report_item_mock.side_effect = Exception("Something is wrong with the license server!")

    assert await license_report.report() == [
        LicenseReportItem(
            feature_id=1,
            product_feature="converge.converge_super",
            used=0,
            total=0,
            uses=[],
        )
    ]


@mark.asyncio
@mock.patch("lm_agent.workload_managers.slurm.cmd_utils.scontrol_show_lic")
@mock.patch("lm_agent.services.license_report.get_cluster_configs_from_backend")
@mock.patch("lm_agent.services.license_report.get_local_license_configurations")
@mock.patch("lm_agent.services.license_report.RLMLicenseServer.get_report_item")
async def test_license_report_empty_on_exception_raised_with_multiple_features(
    get_report_item_mock: mock.MagicMock,
    get_local_license_configurations_mock: mock.MagicMock,
    get_cluster_configs_from_backend_mock: mock.MagicMock,
    show_lic_mock: mock.MagicMock,
):
    """
    Do I get an empty report when an exception is raised?
    """
    get_cluster_configs_from_backend_mock.return_value = []
    show_lic_mock.return_value = ""
    get_local_license_configurations_mock.return_value = [
        ConfigurationSchema(
            id=1,
            name="Converge",
            cluster_client_id="dummy",
            features=[
                FeatureSchema(
                    id=1,
                    name="converge_super",
                    product=ProductSchema(id=1, name="converge"),
                    config_id=1,
                    reserved=100,
                    total=1000,
                    used=93,
                    booked_total=0,
                )
            ],
            license_servers=[
                LicenseServerSchema(id=1, config_id=1, host="licserv0001", port=1234),
                LicenseServerSchema(id=3, config_id=1, host="licserv0003", port=8760),
            ],
            grace_time=60,
            type=LicenseServerType.RLM,
        ),
        ConfigurationSchema(
            id=2,
            name="Converge GUI Polygonica",
            cluster_client_id="dummy",
            features=[
                FeatureSchema(
                    id=2,
                    name="converge_gui_polygonica",
                    product=ProductSchema(id=1, name="converge"),
                    config_id=2,
                    reserved=100,
                    total=1000,
                    used=93,
                    booked_total=0,
                )
            ],
            license_servers=[
                LicenseServerSchema(id=1, config_id=2, host="licserv0001", port=1234),
                LicenseServerSchema(id=3, config_id=2, host="licserv0003", port=8760),
            ],
            grace_time=60,
            type=LicenseServerType.RLM,
        ),
    ]
    get_report_item_mock.side_effect = Exception("Something is wrong with the license server!")

    assert await license_report.report() == [
        LicenseReportItem(
            feature_id=1,
            product_feature="converge.converge_super",
            used=0,
            total=0,
            uses=[],
        ),
        LicenseReportItem(
            feature_id=2,
            product_feature="converge.converge_gui_polygonica",
            used=0,
            total=0,
            uses=[],
        ),
    ]


@mark.asyncio
@mock.patch("lm_agent.services.license_report.report")
async def test__update_features__report_empty(report_mock):
    """
    Check the correct behavior when the report is empty in reconcile.
    """
    report_mock.return_value = []
    with raises(LicenseManagerEmptyReportError):
        await license_report.update_features()


@mark.asyncio
@mark.respx(base_url="http://backend")
@mock.patch("lm_agent.services.license_report.report")
async def test__update_features__put_failed(
    report_mock,
    respx_mock,
):
    """
    Check that when put /features/bulk status_code is not 200, should raise exception.
    """
    report_mock.return_value = [
        LicenseReportItem(
            feature_id=1,
            product_feature="abaqus.abaqus",
            total=1000,
            used=200,
            uses=[],
        )
    ]

    respx_mock.put("/lm/features/bulk").mock(return_value=Response(status_code=400))

    with raises(LicenseManagerBackendConnectionError):
        await license_report.update_features()
